# tests for ccache file
import pathlib
import base64
import tempfile
from minikerberos.common.kirbi import Kirbi
from minikerberos.common.ccache import CCACHE, Header, DateTime, Times,\
	Address, CCACHEOctetString, Authdata, Credential
from .config import *

def test_load_ccache():
	for ccachefile in get_testfiles_ccache():
		ccache = CCACHE.from_file(ccachefile)
		assert len(ccache.credentials) > 0
		targets = [x for x in ccache.list_targets()]
		assert len(targets) > 0
		with tempfile.NamedTemporaryFile() as f:
			ccache.to_file(f.name)
			ccache2 = CCACHE.from_file(f.name)
			targets2 = [x for x in ccache2.list_targets()]
			assert targets == targets2
			assert str(ccache) == str(ccache2)

def test_load_bytes():
	for ccachefile in get_testfiles_ccache():
		with open(ccachefile, 'rb') as f:
			ccache_bytes = f.read()
			ccache = CCACHE.from_bytes(ccache_bytes)
		assert len(ccache.credentials) > 0
		targets = [x for x in ccache.list_targets()]
		assert len(targets) > 0
		ccache_bytes = ccache.to_bytes()
		ccache2 = CCACHE.from_bytes(ccache_bytes)
		targets2 = [x for x in ccache2.list_targets()]
		assert listcompare(targets, targets2)

def test_load_hex():
	for ccachefile in get_testfiles_ccache():
		ccache = CCACHE.from_file(ccachefile)
		assert len(ccache.credentials) > 0
		targets = [x for x in ccache.list_targets()]
		assert len(targets) > 0
		ccache_hex = ccache.to_hex()
		ccache2 = CCACHE.from_hex(ccache_hex)
		targets2 = [x for x in ccache2.list_targets()]
		assert listcompare(targets, targets2)

def test_load_b64():
	for ccachefile in get_testfiles_ccache():
		ccache = CCACHE.from_file(ccachefile)
		assert len(ccache.credentials) > 0
		targets = [x for x in ccache.list_targets()]
		assert len(targets) > 0
		ccache_b64 = ccache.to_b64()
		ccache2 = CCACHE.from_b64(ccache_b64)
		targets2 = [x for x in ccache2.list_targets()]
		assert listcompare(targets, targets2)

def test_to_kirbidir():
	for ccachefile in get_testfiles_ccache():
		ccache = CCACHE.from_file(ccachefile)
		assert len(ccache.credentials) > 0
		targets = [x for x in ccache.list_targets()]
		assert len(targets) > 0
		with tempfile.TemporaryDirectory() as d:
			ccache.to_kirbidir(d)
			ccache2 = CCACHE.from_kirbidir(d)
			targets2 = [x for x in ccache2.list_targets()]
			assert listcompare(targets, targets2)

def test_from_kirbidir():
	ccache = CCACHE.from_kirbidir(KIRBI_DIR)
	assert len(ccache.credentials) > 0
	targets = [x for x in ccache.list_targets()]
	assert len(targets) > 0
	with tempfile.TemporaryDirectory() as d:
		ccache.to_kirbidir(d)
		ccache2 = CCACHE.from_kirbidir(d)
		targets2 = [x for x in ccache2.list_targets()]
		assert listcompare(targets, targets2)

def test_from_kirbifile():
	for kirbifile in get_testfiles_kirbi():
		ccache = CCACHE.from_kirbifile(kirbifile)
		assert len(ccache.credentials) > 0
		targets = [x for x in ccache.list_targets()]
		assert len(targets) > 0

def test_stupid_kirbifile():
	for kirbifile in get_testfiles_kirbi():
		with open(kirbifile, 'rb') as f:
			kirbifile = base64.b64encode(f.read())
		with tempfile.NamedTemporaryFile() as f:
			f.write(kirbifile)
			f.flush()
			ccache = CCACHE.from_kirbifile(f.name)
			assert len(ccache.credentials) > 0
			targets = [x for x in ccache.list_targets()]
			assert len(targets) > 0
		
	
def test_get_hashes():
	ccache = CCACHE.from_kirbidir(KIRBI_DIR)
	assert len(ccache.credentials) > 0
	hashes = ccache.get_hashes()
	assert len(hashes) > 0

def test_get_all_tgt():
	for ccachefile in get_testfiles_ccache():
		if ccachefile.name != 'administrator.ccache':
			continue
		ccache = CCACHE.from_file(ccachefile)
		assert len(ccache.credentials) > 0
		tgt = ccache.get_all_tgt()
		assert len(tgt) == 1
		tgt, key = tgt[0]
		assert tgt['pvno'] == 5
		assert tgt['msg-type'] == 11
		assert tgt['crealm'] == 'POUDLARD.WIZARD'
		assert tgt['cname']['name-type'] == 1
		assert tgt['cname']['name-string'] == ['administrator']
		assert key['keytype'] == 18
		assert len(key['keyvalue']) == 32


def test_get_all_tgt_1():
	# TODO: add a test case with multiple TGSs
	# there should be another test case with actual tgs
	for ccachefile in get_testfiles_ccache():
		if ccachefile.name != 'administrator.ccache':
			continue
		ccache = CCACHE.from_file(ccachefile)
		assert len(ccache.credentials) > 0
		tgs = ccache.get_all_tgs()
		assert len(tgs) == 0



def test_header():
	h = Header()
	h.tag = 1
	h.taglen = 4
	h.tagdata = b'abcd'
	assert h.to_bytes() == b'\x00\x01\x00\x04abcd'
	h2 = Header.from_bytes(h.to_bytes())[0]
	assert h2.tag == 1
	assert h2.taglen == 4
	assert h2.tagdata == b'abcd'
	assert str(h) == str(h2)

def test_datetime():
	dt = DateTime()
	dt.time_offset = 1999323
	dt.usec_offset = 1234567890
	dt2 = DateTime.from_bytes(dt.to_bytes())
	assert dt.time_offset == dt2.time_offset
	assert dt.usec_offset == dt2.usec_offset

def test_times():
	time = Times.dummy_time()
	time2 = Times.from_bytes(time.to_bytes())
	assert time.starttime == time2.starttime
	assert time.endtime == time2.endtime
	assert time.renew_till == time2.renew_till
	assert time.authtime == time2.authtime

def test_address():
	os = CCACHEOctetString()
	os.length = 4
	os.data = b'abcd'

	addr = Address()
	addr.addrtype = 2
	addr.addrdata = os
	addr2 = Address.from_bytes(addr.to_bytes())
	assert addr.addrtype == addr2.addrtype
	assert addr.addrdata == addr2.addrdata

def test_authdata():
	os = CCACHEOctetString()
	os.length = 4
	os.data = b'abcd'

	ad = Authdata()
	ad.authtype = 1
	ad.authdata = os
	ad2 = Authdata.from_bytes(ad.to_bytes())
	assert ad.authtype == ad2.authtype
	assert ad.authdata == ad2.authdata

def test_octetstring_string():
	tdata = 'HELLOWORLD!!!!'
	os = CCACHEOctetString.from_string(tdata)
	assert os.length == len(tdata)
	assert os.data == tdata.encode()

def test_credential_summary():
	for ccachefile in get_testfiles_ccache():
		ccache = CCACHE.from_file(ccachefile)
		assert len(ccache.credentials) > 0
		for cred in ccache.credentials:
			assert cred.summary() != ''


#def test_credential_tgs():
#	for ccachefile in get_testfiles_ccache():
#		ccache = CCACHE.from_file(ccachefile)
#		assert len(ccache.credentials) > 0
#		for cred in ccache.credentials:
#			tgs, key = cred.to_tgs()
#			cred2 = Credential.from_asn1(tgs, key)
#			assert cred2.to_bytes() == cred.to_bytes()

#def test_credential_kirbi():
#	for kirbifile in get_testfiles_kirbi():
#		kirbi = Kirbi.from_file(kirbifile)
#		cred = Credential.from_kirbi(kirbi)
#		kirbi2, _ = cred.to_kirbi()
#		assert kirbi2.kirbiobj.dump() == kirbi.kirbiobj.dump()
#
#if __name__ == '__main__':
#	test_credential_kirbi()


from unicrypto import hashlib
import datetime
from asn1crypto.core import OrderedDict

from minikerberos.protocol.asn1_structs import EncryptedData, krb5_pvno, \
	PrincipalName, PrincipalName, Realm, Checksum, APOptions, Authenticator,\
	Ticket, AP_REQ

from minikerberos.protocol.encryption import _enctype_table
from minikerberos.protocol.constants import NAME_TYPE, MESSAGE_TYPE
from minikerberos.protocol.structures import AuthenticatorChecksum
from minikerberos.gssapi.channelbindings import ChannelBindingsStruct

def construct_apreq_from_tgs_tgt(tgs, sessionkey, tgt, flags = None, seq_number = 0, ap_opts = [], cb_data = None):
	return construct_apreq_from_tgs(
		tgs,
		sessionkey,
		tgt['crealm'],
		tgt['cname'],
		flags,
		seq_number,
		ap_opts,
		cb_data
	)

def construct_apreq_from_tgs(tgs, sessionkey, crealm, cname, flags = None, seq_number = 0, ap_opts = [], cb_data = None):
	return construct_apreq_from_ticket(
		Ticket(tgs['ticket']).dump(),
		sessionkey,
		crealm,
		cname,
		flags,
		seq_number,
		ap_opts,
		cb_data
	)

def construct_apreq_from_ticket(ticket_data, sessionkey, crealm, cname, flags = None, seq_number = 0, ap_opts = [], cb_data = None):
	now = datetime.datetime.now(datetime.timezone.utc)
	authenticator_data = {}
	authenticator_data['authenticator-vno'] = krb5_pvno
	if isinstance(crealm, Realm):
		authenticator_data['crealm'] = crealm
	else:
		authenticator_data['crealm'] = Realm(crealm)
	
	try:
		authenticator_data['cname'] = PrincipalName(cname)
	except:
		if isinstance(cname, PrincipalName):
			authenticator_data['cname'] = cname
		else:
			authenticator_data['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': [cname]})
	
	authenticator_data['cusec'] = now.microsecond
	authenticator_data['ctime'] = now.replace(microsecond=0)
	if flags is not None:

		ac = AuthenticatorChecksum()
		ac.flags = flags
		ac.channel_binding = b'\x00'*16
		if cb_data is not None:
			cb_struct = ChannelBindingsStruct()
			cb_struct.application_data = cb_data
			ac.channel_binding = hashlib.md5(cb_struct.to_bytes()).digest()

		chksum = {}
		chksum['cksumtype'] = 0x8003
		chksum['checksum'] = ac.to_bytes()

		authenticator_data['cksum'] = Checksum(chksum)
		authenticator_data['seq-number'] = seq_number

	cipher = _enctype_table[sessionkey.enctype]
	authenticator_data_enc = cipher.encrypt(sessionkey, 11, Authenticator(authenticator_data).dump(), None)

	ap_req = {}
	ap_req['pvno'] = krb5_pvno
	ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
	ap_req['ticket'] = Ticket.load(ticket_data)
	ap_req['ap-options'] = APOptions(set(ap_opts))
	ap_req['authenticator'] = EncryptedData({'etype': sessionkey.enctype, 'cipher': authenticator_data_enc})
	return AP_REQ(ap_req).dump()


from __future__ import annotations
import getpass
import collections
import base64
import platform
import copy
from typing import List
import os


from unicrypto import hashlib

from minikerberos.common.constants import KerberosSecretType
from minikerberos.protocol.encryption import string_to_key, Enctype
from minikerberos.protocol.constants import EncryptionType
from minikerberos.common.ccache import CCACHE
from minikerberos.common.keytab import Keytab
from minikerberos.common.kirbi import Kirbi
from asn1crypto import cms
from asn1crypto import algos
from minikerberos.protocol.dirtydh import DirtyDH
from oscrypto.asymmetric import rsa_pkcs1v15_sign, load_private_key
from oscrypto.keys import parse_pkcs12, parse_certificate, parse_private

def get_encoded_data(data:bytes or str, encoding = 'file') -> bytes:
	if encoding == 'file':
		with open(data, 'rb') as kf:
			return kf.read()
	elif encoding == 'hex':
		return bytes.fromhex(data)
	elif encoding == 'b64' or encoding == 'base64':
		if isinstance(data, str):
			data = data.encode()
		return base64.b64decode(data)
	elif encoding == 'raw':
		if isinstance(data, str):
			data = data.encode()
		return data
	raise Exception('Unknown encoding "%s"!' % encoding)

class KerberosCredential:
	def __init__(self):
		self.username:str = None
		self.domain:str = None
		self.password:str = None
		self.nt_hash:str = None
		self.lm_hash:str = None
		self.kerberos_key_aes_256:str = None
		self.kerberos_key_aes_128:str = None
		self.kerberos_key_des:str = None
		self.kerberos_key_rc4:str = None
		self.kerberos_key_des3:str = None
		self.certificate = None
		self.private_key = None
		self.__hcert = None #handle on the windows certificate store
		self.__use_windows_certstore:bool = False
		self.commonname:str = None
		self.certstore_name:str = None
		self.dhparams:DirtyDH = None
		self.ccache:CCACHE = None
		self.ccache_spn_strict_check:bool = False
		self.nopreauth = False
		self.override_etypes:List[EncryptionType] = []

	def get_preferred_enctype(self, server_enctypes:List[EncryptionType]) -> EncryptionType:
		client_enctypes = self.get_supported_enctypes(as_int=False)
		common_enctypes = list(set([s_enctype for s_enctype in server_enctypes]) & set(client_enctypes))

		for c_enctype in client_enctypes:
			if c_enctype in common_enctypes:
				return c_enctype

		raise Exception('No common supported enctypes! Server: %s Client: %s' % (
				', '.join([s_enctype.name for s_enctype in server_enctypes]),
				', '.join([c_enctype.name for c_enctype in client_enctypes])
			)
		)

	def get_key_for_enctype(self, etype:EncryptionType, salt:bytes = None) -> bytes:
		"""
		Returns the encryption key bytes for the enctryption type.
		"""
		if etype == EncryptionType.AES256_CTS_HMAC_SHA1_96:
			if self.kerberos_key_aes_256:
				return bytes.fromhex(self.kerberos_key_aes_256)
			if self.password is not None:
				if not salt:
					salt = (self.domain.upper() + self.username).encode()
				return string_to_key(Enctype.AES256, self.password, salt).contents
			raise Exception('There is no key for AES256 encryption')
		elif etype == EncryptionType.AES128_CTS_HMAC_SHA1_96:
			if self.kerberos_key_aes_128:
				return bytes.fromhex(self.kerberos_key_aes_128)
			if self.password is not None:
				if not salt:
					salt = (self.domain.upper() + self.username).encode()
				return string_to_key(Enctype.AES128, self.password, salt).contents
			raise Exception('There is no key for AES128 encryption')
		elif etype == EncryptionType.ARCFOUR_HMAC_MD5:
			if self.kerberos_key_rc4:
				return bytes.fromhex(self.kerberos_key_rc4)
			if self.nt_hash:
				return bytes.fromhex(self.nt_hash)
			elif self.password:
				if isinstance(self.password, str):
					pw = self.password.encode('utf-16-le')
				self.nt_hash = hashlib.md4(pw).hexdigest().upper()
				return bytes.fromhex(self.nt_hash)
			else:
				raise Exception('There is no key for RC4 encryption')
		elif etype == EncryptionType.DES3_CBC_SHA1:
			if self.kerberos_key_des3:
				return bytes.fromhex(self.kerberos_key_des3)
			elif self.password:
				if not salt:
					salt = (self.domain.upper() + self.username).encode()
				return string_to_key(Enctype.DES3, self.password, salt).contents
			else:
				raise Exception('There is no key for DES3 encryption')

		elif etype == EncryptionType.DES_CBC_MD5: #etype == EncryptionType.DES_CBC_CRC or etype == EncryptionType.DES_CBC_MD4 or 
			if self.kerberos_key_des:
				return bytes.fromhex(self.kerberos_key_des)
			elif self.password:
				if not salt:
					salt = (self.domain.upper() + self.username).encode()
				return string_to_key(Enctype.DES_MD5, self.password, salt).contents
			else:
				raise Exception('There is no key for DES encryption')
		
		elif etype == EncryptionType.ARCFOUR_MD4:
			if self.kerberos_key_rc4:
				return bytes.fromhex(self.kerberos_key_rc4)[:8]
			if self.nt_hash:
				return bytes.fromhex(self.nt_hash)[:8]
			elif self.password:
				pw = self.password
				if isinstance(self.password, str):
					pw = self.password.encode('utf-16-le')
				self.nt_hash = hashlib.md4(pw).hexdigest().upper()
				return bytes.fromhex(self.nt_hash)[:8]
			else:
				raise Exception('There is no key for RC4 encryption')

		else:
			raise Exception('Unsupported encryption type: %s' % etype.name)

	def get_supported_enctypes(self, as_int = True) -> List[EncryptionType]:
		"""
		Returns a list of all EncryptionTypes this credentials can use for authentication
		"""
		supp_enctypes = collections.OrderedDict()
		if self.override_etypes is not None and len(self.override_etypes) > 0:
			for etype in self.override_etypes:
				if isinstance(etype, int):
					etype = EncryptionType(etype)
				supp_enctypes[etype] = 1
		
		else:
			if self.nopreauth is True:
				supp_enctypes[EncryptionType.DES_CBC_CRC] = 1
				supp_enctypes[EncryptionType.DES_CBC_MD4] = 1
				supp_enctypes[EncryptionType.DES_CBC_MD5] = 1
				supp_enctypes[EncryptionType.DES3_CBC_SHA1] = 1
				supp_enctypes[EncryptionType.ARCFOUR_HMAC_MD5] = 1
				supp_enctypes[EncryptionType.ARCFOUR_MD4] = 1
				supp_enctypes[EncryptionType.AES256_CTS_HMAC_SHA1_96] = 1
				supp_enctypes[EncryptionType.AES128_CTS_HMAC_SHA1_96] = 1

			if self.kerberos_key_aes_256:
				supp_enctypes[EncryptionType.AES256_CTS_HMAC_SHA1_96] = 1
			if self.kerberos_key_aes_128:
				supp_enctypes[EncryptionType.AES128_CTS_HMAC_SHA1_96] = 1

			if self.password:
				supp_enctypes[EncryptionType.DES_CBC_CRC] = 1
				supp_enctypes[EncryptionType.DES_CBC_MD4] = 1
				supp_enctypes[EncryptionType.DES_CBC_MD5] = 1
				supp_enctypes[EncryptionType.DES3_CBC_SHA1] = 1
				supp_enctypes[EncryptionType.ARCFOUR_HMAC_MD5] = 1
				supp_enctypes[EncryptionType.ARCFOUR_MD4] = 1
				supp_enctypes[EncryptionType.AES256_CTS_HMAC_SHA1_96] = 1
				supp_enctypes[EncryptionType.AES128_CTS_HMAC_SHA1_96] = 1

			if self.password or self.nt_hash or self.kerberos_key_rc4:
				supp_enctypes[EncryptionType.ARCFOUR_HMAC_MD5] = 1
				supp_enctypes[EncryptionType.ARCFOUR_MD4] = 1

			if self.kerberos_key_des:
				supp_enctypes[EncryptionType.DES3_CBC_SHA1] = 1
			
			if self.certificate is not None:
				supp_enctypes = collections.OrderedDict()
				supp_enctypes[EncryptionType.AES256_CTS_HMAC_SHA1_96] = 1
				supp_enctypes[EncryptionType.AES128_CTS_HMAC_SHA1_96] = 1


		if as_int == True:
			return [etype.value for etype in supp_enctypes]
		return [etype for etype in supp_enctypes]

	@staticmethod
	def from_keytab(keytab_file_path: str, principal: str, realm: str, encoding = 'file') -> KerberosCredential:
		"""Returns a kerberos credential object from Keytab file/data"""
		cred = KerberosCredential()
		cred.username = principal
		cred.domain = realm
		data = get_encoded_data(keytab_file_path, encoding=encoding)
		return KerberosCredential.from_keytab_string(data, principal, realm)

	@staticmethod
	def from_ccache(data, principal: str = None, realm: str = None, encoding = 'file') -> KerberosCredential:
		"""Returns a kerberos credential object with CCACHE database"""
		if data is None:
			ccache_path = os.environ.get('KRB5CCNAME')
			if ccache_path is None:
				raise Exception('No CCACHE data or path provided!')
			data = ccache_path
		data = get_encoded_data(data, encoding=encoding)
		k = KerberosCredential()
		k.username = principal
		k.domain = realm
		k.ccache = CCACHE.from_bytes(data)
		return k

	@staticmethod
	def from_kirbi(keytab_file_path: str, principal: str = None, realm: str = None, encoding = 'file') -> KerberosCredential:
		"""Returns a kerberos credential object from .kirbi file"""
		if encoding != 'kirbi':
			data = get_encoded_data(keytab_file_path, encoding=encoding)
			kirbi = Kirbi.from_bytes(data)
		else:
			kirbi = copy.deepcopy(keytab_file_path)
		cred = KerberosCredential()
		cred.username = principal if principal is not None else kirbi.get_username()
		cred.domain = realm if realm is not None else kirbi.kirbiobj.native['tickets'][0]['realm']
		cred.ccache = CCACHE.from_kirbi(kirbi)
		cred.ccache_spn_strict_check = False
		return cred
	
	@staticmethod
	def from_pfx(data:str, password:str, dhparams:DirtyDH = None, username:str = None, domain:str = None, encoding = 'file') -> KerberosCredential:
		"""
		Retruns a credential object from data found in the PFX file
		Username and domain will override the values found in the certificate
		"""
		data = get_encoded_data(data, encoding=encoding)
		return KerberosCredential.from_pfx_string(data, password, dhparams = dhparams, username = username, domain = domain)
	
	@staticmethod
	def from_krbcred(keytab_file_path: str, principal: str = None, realm: str = None) -> KerberosCredential:
		return KerberosCredential.from_kirbi(keytab_file_path, principal, realm)
	
	@staticmethod
	def from_keytab_string(keytabdata: str or bytes, principal: str, realm: str) -> KerberosCredential:
		cred = KerberosCredential()
		if principal is None:
			raise Exception('Principal is required')
		if realm is None:
			raise Exception('Realm is required')
		
		cred.username = principal
		cred.domain = realm

		if isinstance(keytabdata, str):
			keytabdata = base64.b64decode(keytabdata.replace(' ','').replace('\r','').replace('\n','').replace('\t','').replace('','').encode())
		
		keytab = Keytab.from_bytes(keytabdata)
		if len(keytab.entries) == 0:
			raise Exception('No entries found in keytab')
		if len(keytab.entries) == 1:
			keytab_entry = keytab.entries[0]
		else:
			for entry in keytab.entries:
				if entry.principal.to_pname().lower().find(principal.lower()) != -1:
					keytab_entry = entry
					break
			else:
				# No match found
				keytab_entry = keytab.entries[0]

		if Enctype.AES256 == keytab_entry.enctype:
			enctype = KerberosSecretType.AES256
		elif Enctype.AES128 == keytab_entry.enctype:
			enctype = KerberosSecretType.AES128
		elif Enctype.DES3 == keytab_entry.enctype:
			enctype = KerberosSecretType.DES3
		elif Enctype.DES_CRC == keytab_entry.enctype:
			enctype = KerberosSecretType.DES
		elif Enctype.DES_MD4 == keytab_entry.enctype:
			enctype = KerberosSecretType.DES
		elif Enctype.DES_MD5 == keytab_entry.enctype:
			enctype = KerberosSecretType.DES
		elif Enctype.RC4 == keytab_entry.enctype:
			enctype = KerberosSecretType.RC4
		if enctype:
			cred.add_secret(enctype, keytab_entry.key_contents.hex())
		
		return cred

	def set_user_and_domain_from_cert(self, username:str = None, domain:str = None):
		"""
		Tries to guess the correct username and domain from the current certificate,
		if 'username' and/or 'domain' is set it will set those
		"""
		self.username = username
		if username is None:
			self.username = self.certificate.subject.native['common_name'].rsplit('@', 1)[0]
		self.domain = domain
		if domain is None:
			dc = None
			if 'domain_component' in self.certificate.issuer.native:
				dc = self.certificate.issuer.native['domain_component']
			if 'domain_component' in self.certificate.subject.native:
				dc = self.certificate.subject.native['domain_component']
			if dc is not None:
				self.domain = '.'.join(dc[::-1])
			else:
				raise Exception('Could\'t find proper domain name in the certificate! Please set it manually!')

	@staticmethod
	def from_pem_data(certdata: str or bytes, keydata:str or bytes, dhparams:DirtyDH = None, username:str = None, domain:str = None) -> KerberosCredential:
		if isinstance(certdata, str):
			certdata = base64.b64decode(certdata.replace(' ','').replace('\r','').replace('\n','').replace('\t',''))
		if isinstance(keydata, str):
			keydata = base64.b64decode(keydata.replace(' ','').replace('\r','').replace('\n','').replace('\t',''))
		k = KerberosCredential()
		k.certificate = parse_certificate(certdata)
		k.private_key = parse_private(keydata)
		k.set_user_and_domain_from_cert(username = username, domain = domain)
		k.set_dhparams(dhparams)
		return k

	@staticmethod
	def from_pem_file(certpath:str, keypath: str, dhparams:DirtyDH = None, username:str = None, domain:str = None) -> KerberosCredential:
		
		with open(certpath, 'rb') as f:
			certdata = f.read()

		with open(keypath, 'rb') as f:
			keydata = f.read()
		
		return KerberosCredential.from_pem_data(certdata, keydata, dhparams = dhparams, username = username, domain = domain)


	@staticmethod
	def from_windows_certstore(commonname:str, certstore_name:str = 'MY', dhparams:DirtyDH = None, username:str = None, domain:str = None) -> KerberosCredential:
		if platform.system().lower() != 'windows':
			raise Exception('Only works on windows (obviously)')
		from minikerberos.common.windows.crypt32 import find_cert_by_cn, CertCloseStore, CertFreeCertificateContext

		k = KerberosCredential()
		k.commonname = commonname
		k.certstore_name = certstore_name
		k.certificate, chandle, shandle = find_cert_by_cn(commonname, certstore_name)
		CertFreeCertificateContext(chandle)
		CertCloseStore(shandle)

		k.__use_windows_certstore = True
		k.set_user_and_domain_from_cert(username = username, domain = domain)
		k.set_dhparams(dhparams)
		return k

	@staticmethod
	def from_pfx_string(data: str or bytes, password:str, dhparams:DirtyDH = None, username:str = None, domain:str = None) -> KerberosCredential:
		k = KerberosCredential()
		if password is None:
			password = b''
		if isinstance(password, str):
			password = password.encode()
		
		if isinstance(data, str):
			data = base64.b64decode(data.replace(' ', '').replace('\r','').replace('\n','').encode())

		# private_key is not actually the private key object but the privkey data because oscrypto privkey 
		# cant be serialized so we cant make copy of it.
		k.private_key, k.certificate, extra_certs = parse_pkcs12(data, password = password)
		#k.private_key = load_private_key(privkeyinfo)
		
		k.set_user_and_domain_from_cert(username = username, domain = domain)
		k.set_dhparams(dhparams)
		return k

	@staticmethod
	def from_pfx_file(filepath:str, password:str, dhparams:DirtyDH = None, username:str = None, domain:str = None) -> KerberosCredential:
		"""
		Username and domain will override the values found in the certificate
		"""
		with open(filepath, 'rb') as f:
			data = f.read()
		return KerberosCredential.from_pfx_string(data, password, dhparams = dhparams, username = username, domain = domain)
	
	def set_dhparams(self, dhparams):
		# windows default params, don't look at me...
		self.dhparams = DirtyDH.from_dict({
			'p':int('00ffffffffffffffffc90fdaa22168c234c4c6628b80dc1cd129024e088a67cc74020bbea63b139b22514a08798e3404ddef9519b3cd3a431b302b0a6df25f14374fe1356d6d51c245e485b576625e7ec6f44c42e9a637ed6b0bff5cb6f406b7edee386bfb5a899fa5ae9f24117c4b1fe649286651ece65381ffffffffffffffff', 16),
			'g':2
		})
		
		if dhparams is not None:
			if isinstance(dhparams, dict):
				self.dhparams = DirtyDH.from_dict(dhparams)
			elif isinstance(dhparams, bytes):
				self.dhparams = DirtyDH.from_asn1(dhparams)
			elif isinstance(dhparams, DirtyDH):
				self.dhparams= dhparams
			else:
				raise Exception('DH params must be either a bytearray or a dict')
		

	def sign_authpack(self, data, wrap_signed = False):
		if self.__use_windows_certstore is True:
			from minikerberos.common.windows.crypt32 import pkcs7_sign, CertCloseStore, find_cert_by_cn, CertFreeCertificateContext
			_, chandle, shandle  = find_cert_by_cn(self.commonname, self.certstore_name)
			res = pkcs7_sign(chandle, data)
			CertFreeCertificateContext(chandle)
			CertCloseStore(shandle)
			return res
		return self.sign_authpack_native(data, wrap_signed)

	def sign_authpack_native(self, data, wrap_signed = False):
		"""
		Creating PKCS7 blob which contains the following things:

		1. 'data' blob which is an ASN1 encoded "AuthPack" structure
		2. the certificate used to sign the data blob
		3. the singed 'signed_attrs' structure (ASN1) which points to the "data" structure (in point 1)
		"""
		
		da = {}
		da['algorithm'] = algos.DigestAlgorithmId('1.3.14.3.2.26') # for sha1

		si = {}
		si['version'] = 'v1'
		si['sid'] = cms.IssuerAndSerialNumber({
			'issuer':  self.certificate.issuer,
			'serial_number':  self.certificate.serial_number,
		})


		si['digest_algorithm'] = algos.DigestAlgorithm(da)
		si['signed_attrs'] = [
			cms.CMSAttribute({'type': 'content_type', 'values': ['1.3.6.1.5.2.3.1']}), # indicates that the encap_content_info's authdata struct (marked with OID '1.3.6.1.5.2.3.1' is signed )
			cms.CMSAttribute({'type': 'message_digest', 'values': [hashlib.sha1(data).digest()]}), ### hash of the data, the data itself will not be signed, but this block of data will be.
		]
		si['signature_algorithm'] = algos.SignedDigestAlgorithm({'algorithm' : '1.2.840.113549.1.1.1'})
		si['signature'] = rsa_pkcs1v15_sign(load_private_key(self.private_key),  cms.CMSAttributes(si['signed_attrs']).dump(), "sha1")

		ec = {}
		ec['content_type'] = '1.3.6.1.5.2.3.1'
		ec['content'] = data

		sd = {}
		sd['version'] = 'v3'
		sd['digest_algorithms'] = [algos.DigestAlgorithm(da)] # must have only one
		sd['encap_content_info'] = cms.EncapsulatedContentInfo(ec)
		sd['certificates'] = [self.certificate]
		sd['signer_infos'] = cms.SignerInfos([cms.SignerInfo(si)])
		
		if wrap_signed is True:
			ci = {}
			ci['content_type'] = '1.2.840.113549.1.7.2' # signed data OID
			ci['content'] = cms.SignedData(sd)
			return cms.ContentInfo(ci).dump()

		return cms.SignedData(sd).dump()

	def add_secret(self, st: KerberosSecretType, secret: str):
		if st == KerberosSecretType.PASSWORD or st == KerberosSecretType.PW or st == KerberosSecretType.PASS:
			if secret == '' or secret is None:
				self.password = getpass.getpass('Enter Kerberos credential password:')
			else:
				self.password = secret
		elif st == KerberosSecretType.NT or st == KerberosSecretType.RC4:
			self.nt_hash = secret
			self.kerberos_key_rc4 = secret
		elif st == KerberosSecretType.AES128:
			self.kerberos_key_aes_128 = secret
		elif st == KerberosSecretType.AES256:
			self.kerberos_key_aes_256 = secret
		elif st == KerberosSecretType.AES:
			bytes.fromhex(secret)
			if len(secret) == 32:
				self.kerberos_key_aes_128 = secret
			elif len(secret) == 64:
				self.kerberos_key_aes_256 = secret
			else:
				raise Exception('AES key incorrect length!')
		elif st == KerberosSecretType.DES:
			self.kerberos_key_des = secret
		elif st == KerberosSecretType.DES3 or st == KerberosSecretType.TDES:
			self.kerberos_key_des3 = secret
		elif st == KerberosSecretType.CCACHE:
			self.ccache = CCACHE.from_file(secret)

	@staticmethod
	def from_tgt(tgt, override_realm = None, override_etypes:List[int] = [17,18,23]):
		"""Returns a new KerberosCredential object with user matching the TGT"""
		new_cred = KerberosCredential()
		new_cred.username = tgt['cname']['name-string'][0]
		if len(tgt['cname']['name-string']) > 1:
			new_cred.domain = tgt['cname']['name-string'][1]
		new_cred.domain = tgt['crealm']
		if override_realm is not None:
			new_cred.domain = override_realm
		if override_etypes is not None:
			for etype in override_etypes:
				new_cred.override_etypes.append(EncryptionType(etype))
		return new_cred

	def __str__(self):
		t = '===KerberosCredential===\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				t += '%s: %s\r\n' % (k, ','.join([str(x) for x in self.__dict__[k]]))
			else:
				t += '%s: %s\r\n' % (k, self.__dict__[k])
		return t
		



import os
import logging
import asyncio
import copy
from minikerberos.common.factory import KerberosClientFactory, kerberos_url_help_epilog
from minikerberos.common.target import KerberosTarget
from asysocks.unicomm.common.target import UniProto
from minikerberos.common.spn import KerberosSPN
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.common.kirbi import Kirbi
from minikerberos.common.creds import KerberosCredential
from minikerberos import logger

async def getTGS(kerberos_url, kirbifile = None):
	if isinstance(spn, str):
		spn = KerberosSPN.from_spn(spn)

	cu = KerberosClientFactory.from_url(kerberos_url)
	client = cu.get_client()
	logging.debug('Getting TGT')
	await client.get_TGT()
	logging.debug('Getting TGS for otherdomain krbtgt')
	ref_tgs, ref_encpart, ref_key, new_factory = await client.get_referral_ticket(spn.domain)
	kirbi = Kirbi.from_ticketdata(ref_tgs, ref_encpart)
	print(str(kirbi))
	if kirbifile is not None:
		kirbi.to_file(kirbifile)
	
	logging.info('Done!')

def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Polls the kerberos service for a TGS for the sepcified user and specified service', formatter_class=argparse.RawDescriptionHelpFormatter, epilog = kerberos_url_help_epilog)
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('--kirbi', help='kirbi file to store the TGT ticket in, otherwise kirbi will be printed to stdout')
	parser.add_argument('kerberos_url', help='the kerberos target string. ')

	logger.setLevel(logging.DEBUG)
	args = parser.parse_args()
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig(level=1)
	
	asyncio.run(getTGS(args.kerberos_url, args.kirbi))
	
if __name__ == '__main__':
	main()


from minikerberos.protocol.external.ndr import NDRSTRUCT, NDRUniConformantArray, NDRPOINTER, NDRUniFixedArray
from minikerberos.protocol.external.dtypes import RPC_UNICODE_STRING, ULONG


# 2.2.1.2.2 NL_SITE_NAME_ARRAY
class RPC_UNICODE_STRING_ARRAY(NDRUniConformantArray):
    item = RPC_UNICODE_STRING

class PRPC_UNICODE_STRING_ARRAY(NDRPOINTER):
    referent = (
        ('Data', RPC_UNICODE_STRING_ARRAY),
    )

class UCHAR_ARRAY(NDRUniConformantArray):
    item = 'c'

class PUCHAR_ARRAY(NDRPOINTER):
    referent = (
        ('Data', UCHAR_ARRAY),
    )

class PRPC_UNICODE_STRING_ARRAY(NDRPOINTER):
    referent = (
        ('Data', RPC_UNICODE_STRING_ARRAY),
    )


class PUCHAR_ARRAY(NDRPOINTER):
    referent = (
        ('Data', UCHAR_ARRAY),
    )


# 2.2.1.1.1 CYPHER_BLOCK
class CYPHER_BLOCK(NDRSTRUCT):
    structure = (
        ('Data', '8s=b""'),
    )
    def getAlignment(self):
        return 1
        


# 2.2.1.1.3 LM_OWF_PASSWORD
class CYPHER_BLOCK_ARRAY(NDRUniFixedArray):
    def getDataLen(self, data, offset=0):
        return len(CYPHER_BLOCK())*2

class LM_OWF_PASSWORD(NDRSTRUCT):
    structure = (
        ('Data', CYPHER_BLOCK_ARRAY),
    )

# 2.2.1.1.4 NT_OWF_PASSWORD
NT_OWF_PASSWORD = LM_OWF_PASSWORD
ENCRYPTED_NT_OWF_PASSWORD = NT_OWF_PASSWORD


# 2.2.1.4.1 LM_CHALLENGE
class CHAR_FIXED_8_ARRAY(NDRUniFixedArray):
    def getDataLen(self, data, offset=0):
        return 8

# 2.2.1.4.9 USER_SESSION_KEY
USER_SESSION_KEY = LM_OWF_PASSWORD

# 2.2.1.4.10 GROUP_MEMBERSHIP
class GROUP_MEMBERSHIP(NDRSTRUCT):
    structure = (
        ('RelativeId', ULONG),
        ('Attributes', ULONG),
    )

class GROUP_MEMBERSHIP_ARRAY(NDRUniConformantArray):
    item = GROUP_MEMBERSHIP

class PGROUP_MEMBERSHIP_ARRAY(NDRPOINTER):
    referent = (
        ('Data', GROUP_MEMBERSHIP_ARRAY),
    )


from minikerberos import logger
from minikerberos.protocol.asn1_structs import KerberosResponse
from minikerberos.common.target import KerberosTarget
from asysocks.unicomm.common.target import UniProto
from asysocks.unicomm.client import UniClient
from asysocks.unicomm.common.packetizers import Packetizer


class KerberosPacketizer(Packetizer):
	def __init__(self, buffer_size = 65535):
		Packetizer.__init__(self, buffer_size)
		self.buffer_size = buffer_size
		self.in_buffer = b''
	
	def process_buffer(self):
		if len(self.in_buffer) > 4:
			length = int.from_bytes(self.in_buffer[:4], byteorder = 'big', signed = False)
			if len(self.in_buffer) >= length:
				data = self.in_buffer[4:4+length]
				self.in_buffer = self.in_buffer[length+4:]
				yield data
				
	async def data_out(self, data):
		yield data

	async def data_in(self, data):
		if data is None:
			yield data
		self.in_buffer += data
		for packet in self.process_buffer():
			yield packet

class AIOKerberosClientSocket:
	def __init__(self, target:KerberosTarget):
		self.target = target
	
	def get_addr_str(self):
		return '%s:%d' % (self.target.get_hostname_or_ip(), self.target.port)
	
	async def sendrecv(self, data, throw:bool = False):
		client = None
		connection = None
		try:
			packetizer = KerberosPacketizer()
			client = UniClient(self.target, packetizer)
			connection = await client.connect()
			if self.target.protocol == UniProto.CLIENT_TCP:
				length = len(data).to_bytes(4, byteorder = 'big', signed = False)
				await connection.write(length + data)
				
				async for packet in connection.read():
					krb_message = KerberosResponse.load(packet)
					break
				
			elif self.target.protocol == UniProto.CLIENT_UDP:
				raise Exception('Not implemented!')
			
			return krb_message
		finally:
			if connection is not None:
				await connection.close()




from minikerberos.common.ccache import CCACHE

def ccacheroast(ccachefilepath):
	ccache = CCACHE.from_file(ccachefilepath)
	for hash in ccache.get_hashes():
		print(hash)

def main():
	import argparse
	parser = argparse.ArgumentParser(description='Parses CCACHE file and outputs all TGS tickets in a hashcat-crackable format')
	parser.add_argument('ccache', help='CCACHE file to roast')
	
	args = parser.parse_args()
	ccacheroast(args.ccache)
	
		
		
if __name__ == '__main__':
	main()


import collections
import datetime
import secrets
from typing import List, Dict
from unicrypto import hashlib

from minikerberos import logger
from minikerberos.common.ccache import CCACHE
from minikerberos.network.clientsocket import KerberosClientSocket
from minikerberos.protocol.asn1_structs import METHOD_DATA, ETYPE_INFO, ETYPE_INFO2, \
	PADATA_TYPE, PA_PAC_REQUEST, PA_ENC_TS_ENC, EncryptedData, krb5_pvno, KDC_REQ_BODY, \
	AS_REQ, TGS_REP, KDCOptions, PrincipalName, EncASRepPart, EncTGSRepPart, PrincipalName, Realm, \
	Checksum, APOptions, Authenticator, Ticket, AP_REQ, TGS_REQ, CKSUMTYPE, \
	PA_FOR_USER_ENC, PA_PAC_OPTIONS, PA_PAC_OPTIONSTypes, EncTicketPart

from minikerberos.protocol.errors import KerberosErrorCode, KerberosError
from minikerberos.protocol.encryption import Key, _enctype_table, _HMACMD5, Enctype
from minikerberos.protocol.constants import PaDataType, EncryptionType, NAME_TYPE, MESSAGE_TYPE
from minikerberos.protocol.structures import AuthenticatorChecksum
from minikerberos.protocol.rfc4556 import PKAuthenticator, AuthPack, PA_PK_AS_REP, KDCDHKeyInfo, PA_PK_AS_REQ
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.target import KerberosTarget
from minikerberos.common.spn import KerberosSPN
from minikerberos.protocol.ticketutils import construct_apreq_from_tgs_tgt


from asn1crypto import cms
from asn1crypto import core

class KerbrosClient:
	def __init__(self, ccred:KerberosCredential, target:KerberosTarget):
		self.credential = ccred
		self.target = target
		self.ksoc = KerberosClientSocket(self.target)
		self.ccache = CCACHE() if ccred.ccache is None else ccred.ccache
		self.kerberos_session_key = None
		self.kerberos_TGT = None
		self.kerberos_TGT_encpart = None
		self.kerberos_TGS = None
		self.kerberos_cipher = None
		self.kerberos_cipher_type = None
		self.kerberos_key = None
		self.server_salt = None
		self.pkinit_tkey = None

	def build_asreq_lts(self, supported_encryption_method, kdcopts:List[str] = ['forwardable','renewable','proxiable'], enctimestamp=None, newnow=None, no_preauth = False, kdc_req_body_extra = None, with_pac:bool = True) -> AS_REQ:
		logger.debug('Constructing TGT request with auth data')
		#now to create an AS_REQ with encrypted timestamp for authentication
		padatas = []
		
		if with_pac is True:
			pa_data_1 = {}
			pa_data_1['padata-type'] = int(PADATA_TYPE('PA-PAC-REQUEST'))
			pa_data_1['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()
			padatas.append(pa_data_1)
		
		logger.debug('Selecting common encryption type: %s' % supported_encryption_method.name)
		now = datetime.datetime.now(datetime.timezone.utc)
		if no_preauth is False:
			if enctimestamp is None:
				#creating timestamp asn1
				timestamp = PA_ENC_TS_ENC({'patimestamp': now.replace(microsecond=0), 'pausec': now.microsecond}).dump()
				self.kerberos_cipher = _enctype_table[supported_encryption_method.value]
				self.kerberos_cipher_type = supported_encryption_method.value
				self.kerberos_key = Key(self.kerberos_cipher.enctype, self.credential.get_key_for_enctype(supported_encryption_method, salt = self.server_salt))
				enc_timestamp = self.kerberos_cipher.encrypt(self.kerberos_key, 1, timestamp, None)
			else:
				now = newnow
				enc_timestamp = enctimestamp
			
			pa_data_2 = {}
			pa_data_2['padata-type'] = int(PADATA_TYPE('ENC-TIMESTAMP'))
			pa_data_2['padata-value'] = EncryptedData({'etype': supported_encryption_method.value, 'cipher': enc_timestamp}).dump()
			padatas.append(pa_data_2)

		kdc_req_body = {}
		kdc_req_body['kdc-options'] = KDCOptions(set(kdcopts))
		kdc_req_body['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': self.credential.username.split('/')})
		kdc_req_body['realm'] = self.credential.domain.upper()
		kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': ['krbtgt', self.credential.domain.upper()]})
		kdc_req_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['nonce'] = secrets.randbits(31)
		kdc_req_body['etype'] = [supported_encryption_method.value] #selecting according to server's preferences
		
		if kdc_req_body_extra is not None:
			for key in kdc_req_body_extra:
				kdc_req_body[key] = kdc_req_body_extra[key]
		
		kdc_req = {}
		kdc_req['pvno'] = krb5_pvno
		kdc_req['msg-type'] = MESSAGE_TYPE.KRB_AS_REQ.value
		kdc_req['padata'] = padatas
		kdc_req['req-body'] = KDC_REQ_BODY(kdc_req_body)
		
		return AS_REQ(kdc_req)
	
	def build_asreq_pkinit(self, supported_encryption_method, kdcopts = ['forwardable','renewable','renewable-ok'], with_pac:bool = True) -> AS_REQ:
		from asn1crypto import keys

		if supported_encryption_method.value == 23:
			raise Exception('RC4 encryption is not supported for certificate auth!')


		now = datetime.datetime.now(datetime.timezone.utc)

		kdc_req_body_data = {}
		kdc_req_body_data['kdc-options'] = KDCOptions(set(kdcopts))
		kdc_req_body_data['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': self.credential.username.split('/')})
		kdc_req_body_data['realm'] = self.credential.domain.upper()
		kdc_req_body_data['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': ['krbtgt', self.credential.domain.upper()]})
		kdc_req_body_data['till']  = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body_data['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body_data['nonce'] = secrets.randbits(31)
		kdc_req_body_data['etype'] = [supported_encryption_method.value] #[18,17] # 23 breaks...
		kdc_req_body = KDC_REQ_BODY(kdc_req_body_data)


		checksum = hashlib.sha1(kdc_req_body.dump()).digest()

		authenticator = {}
		authenticator['cusec'] = now.microsecond
		authenticator['ctime'] = now.replace(microsecond=0)
		authenticator['nonce'] = secrets.randbits(31)
		authenticator['paChecksum'] = checksum


		dp = {}
		dp['p'] = self.credential.dhparams.p
		dp['g'] = self.credential.dhparams.g
		dp['q'] = 0 # mandatory parameter, but it is not needed

		pka = {}
		pka['algorithm'] = '1.2.840.10046.2.1'
		pka['parameters'] = keys.DomainParameters(dp)

		spki = {}
		spki['algorithm'] = keys.PublicKeyAlgorithm(pka)
		spki['public_key'] = self.credential.dhparams.get_public_key()


		authpack = {}
		authpack['pkAuthenticator'] = PKAuthenticator(authenticator)
		authpack['clientPublicValue'] = keys.PublicKeyInfo(spki)
		authpack['clientDHNonce'] = self.credential.dhparams.dh_nonce

		authpack = AuthPack(authpack)
		signed_authpack = self.credential.sign_authpack(authpack.dump(), wrap_signed = True)

		payload = PA_PK_AS_REQ()
		payload['signedAuthPack'] = signed_authpack

		padatas = []
		if with_pac is True:
			pa_data_0 = {}
			pa_data_0['padata-type'] = int(PADATA_TYPE('PA-PAC-REQUEST'))
			pa_data_0['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()
			padatas.append(pa_data_0)
		
		pa_data_1 = {}
		pa_data_1['padata-type'] = PaDataType.PK_AS_REQ.value
		pa_data_1['padata-value'] = payload.dump()
		padatas.append(pa_data_1)

		asreq = {}
		asreq['pvno'] = 5
		asreq['msg-type'] = 10
		asreq['padata'] = padatas
		asreq['req-body'] = kdc_req_body

		return AS_REQ(asreq)

	def do_preauth(self, supported_encryption_method, kdcopts = ['forwardable','renewable','renewable-ok'], with_pac:bool = True):
		if self.credential.certificate is not None:
			req = self.build_asreq_pkinit(supported_encryption_method, kdcopts, with_pac=with_pac)
		else:
			req = self.build_asreq_lts(supported_encryption_method, kdcopts, with_pac=with_pac)

		
		logger.debug('Sending TGT request to server')
		rep = self.ksoc.sendrecv(req.dump())
		if rep.name == 'KRB_ERROR':
			raise KerberosError(rep, 'Preauth failed!')
		return rep

	def tgt_from_ccache(self):
		try:
			if self.ccache is None:
				raise Exception('No CCACHE file found')
			
			tgt, keystruct, err = self.ccache.get_tgt(self.credential.username, self.credential.domain, self.credential.ccache_spn_strict_check)
			if err is not None:
				raise err
			self.kerberos_TGT = tgt
			self.kerberos_TGT_encpart = tgt['enc-part']
			self.kerberos_session_key = Key(keystruct['keytype'], keystruct['keyvalue'])
			self.kerberos_cipher = _enctype_table[keystruct['keytype']]
			self.kerberos_cipher_type = keystruct['keytype']
			return True, None
		except Exception as e:
			return None, e
	
	def select_preferred_encryption_method(self, rep):
		#now getting server's supported encryption methods
		
		supp_enc_methods = collections.OrderedDict()
		for enc_method in METHOD_DATA.load(rep['e-data']).native:
			data_type = PaDataType(enc_method['padata-type'])
			
			if data_type == PaDataType.ETYPE_INFO or data_type == PaDataType.ETYPE_INFO2:
				if data_type == PaDataType.ETYPE_INFO:
					enc_info_list = ETYPE_INFO.load(enc_method['padata-value'])
					
				elif data_type == PaDataType.ETYPE_INFO2:
					enc_info_list = ETYPE_INFO2.load(enc_method['padata-value'])
		
				for enc_info in enc_info_list.native:
					supp_enc_methods[EncryptionType(enc_info['etype'])] = enc_info['salt']
					logger.debug('Server supports encryption type %s with salt %s' % (EncryptionType(enc_info['etype']).name, enc_info['salt']))
		
		preferred_enc_type = self.credential.get_preferred_enctype(supp_enc_methods)
		if preferred_enc_type not in supp_enc_methods:
			raise Exception('Preferred enc type not in supported enctypes')
		salt = supp_enc_methods[preferred_enc_type]
		if salt is not None:
			salt = salt.encode()
		self.server_salt = salt #enc_info['salt'].encode()
		return preferred_enc_type

	def get_TGT(self, override_etype = None, decrypt_tgt = True, kdcopts = ['forwardable','renewable','proxiable'], override_sname:KerberosSPN = None, with_pac:bool = True):
		"""
		decrypt_tgt: used for asreproast attacks
		Steps performed:
			1. Send and empty (no encrypted timestamp) AS_REQ with all the encryption types we support
			2. Depending on the response (either error or AS_REP with TGT) we either send another AS_REQ with the encrypted data or return the TGT (or fail miserably)
			3. PROFIT
		"""

		#first, let's check if CCACHE has the correct ticket already
		_, err = self.tgt_from_ccache()
		if err is None:
			return
		
		logger.debug('Generating initial TGT without authentication data')
		now = datetime.datetime.now(datetime.timezone.utc)
		kdc_req_body = {}
		kdc_req_body['kdc-options'] = KDCOptions(set(kdcopts))
		kdc_req_body['cname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': self.credential.username.split('/')})
		kdc_req_body['realm'] = self.credential.domain.upper()
		if override_sname is None:
			kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': ['krbtgt', self.credential.domain.upper()]})
		else:
			# if we want to directly kerberoast with no-preauth user
			kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': override_sname.get_principalname()})
		kdc_req_body['till']  = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['nonce'] = secrets.randbits(31)
		if override_etype is None:
			kdc_req_body['etype'] = self.credential.get_supported_enctypes()
		else:
			kdc_req_body['etype'] = override_etype

		pa_data_1 = {}
		if with_pac is True:
			pa_data_1['padata-type'] = int(PADATA_TYPE('PA-PAC-REQUEST'))
			pa_data_1['padata-value'] = PA_PAC_REQUEST({'include-pac': True}).dump()
		
		kdc_req = {}
		kdc_req['pvno'] = krb5_pvno
		kdc_req['msg-type'] = MESSAGE_TYPE.KRB_AS_REQ.value
		if len(pa_data_1) > 0:
			kdc_req['padata'] = [pa_data_1]
		kdc_req['req-body'] = KDC_REQ_BODY(kdc_req_body)
		
		req = AS_REQ(kdc_req)	
		
		logger.debug('Sending initial TGT to %s' % self.ksoc.get_addr_str())
		rep = self.ksoc.sendrecv(req.dump(), throw = False)

		if rep.name != 'KRB_ERROR':
			#user can do kerberos auth without preauthentication!
			rep = rep.native
			self.kerberos_TGT = rep

			#if we want to roast the asrep (tgt rep) part then we dont even have the proper keys to decrypt
			#so we just return, the asrep can be extracted from this object anyhow
			if decrypt_tgt == False or self.credential.nopreauth is True:
				return rep

			self.kerberos_cipher = _enctype_table[rep['enc-part']['etype']]
			self.kerberos_cipher_type = rep['enc-part']['etype']
			self.kerberos_key = Key(self.kerberos_cipher.enctype, self.credential.get_key_for_enctype(EncryptionType(rep['enc-part']['etype'])))
			
		else:
			if rep.native['error-code'] != KerberosErrorCode.KDC_ERR_PREAUTH_REQUIRED.value:
				raise KerberosError(rep)
			rep = rep.native
			logger.debug('Got reply from server, asikg to provide auth data')
			supported_encryption_method = self.select_preferred_encryption_method(rep)

			rep = self.do_preauth(supported_encryption_method)
			logger.debug('Got valid TGT response from server')
			rep = rep.native
			self.kerberos_TGT = rep

		if self.credential.certificate is not None:
			self.kerberos_TGT_encpart, self.kerberos_session_key, self.kerberos_cipher = self.decrypt_asrep_cert(rep)
			self.kerberos_cipher_type = supported_encryption_method.value

			self.ccache.add_tgt(self.kerberos_TGT, self.kerberos_TGT_encpart, override_pp = True)
			logger.debug('Got valid TGT')
			return 
		
		else:
			cipherText = rep['enc-part']['cipher']
			temp = self.kerberos_cipher.decrypt(self.kerberos_key, 3, cipherText)
			try:
				self.kerberos_TGT_encpart = EncASRepPart.load(temp).native
			except Exception as e:
				logger.debug('EncAsRepPart load failed, is this linux?')
				try:
					self.kerberos_TGT_encpart = EncTGSRepPart.load(temp).native
				except Exception as e:
					logger.error('Failed to load decrypted part of the reply!')
					raise e
			
			self.kerberos_session_key = Key(self.kerberos_cipher.enctype, self.kerberos_TGT_encpart['key']['keyvalue'])
			self.ccache.add_tgt(self.kerberos_TGT, self.kerberos_TGT_encpart, override_pp = True)
			logger.debug('Got valid TGT')
			
			return 
	
	def tgs_from_ccache(self, spn_user:KerberosSPN):
		try:
			if self.ccache is None:
				raise Exception('No CCACHE file found')
			
			tgs, keystruct, err = self.ccache.get_tgs(spn_user)
			if err is not None:
				raise err
			
			key = Key(keystruct['keytype'], keystruct['keyvalue'])
			tgs = TGS_REP(tgs).native
			return tgs, tgs['enc-part'], key, None
		except Exception as e:
			return None, None, None, e
		
	def get_TGS(self, spn_user, override_etype = None, is_linux = False):
		"""
		Requests a TGS ticket for the specified user.
		Retruns the TGS ticket, end the decrpyted encTGSRepPart.

		spn_user: KerberosTarget: the service user you want to get TGS for.
		override_etype: None or list of etype values (int) Used mostly for kerberoasting, will override the AP_REQ supported etype values (which is derived from the TGT) to be able to recieve whatever tgs tiecket 
		"""
		
		#first, let's check if CCACHE has the correct ticket already
		tgs, encTGSRepPart, key, err = self.tgs_from_ccache(spn_user)
		if err is None:
			return tgs, encTGSRepPart, key

		
		if self.kerberos_TGT is None:
			#let's check if CCACHE has a TGT for us
			_, err = self.tgt_from_ccache()
			if err is not None:
				raise Exception('No TGT found in CCACHE!')

		#nope, we need to contact the server
		logger.debug('Constructing TGS request for user %s' % spn_user.get_formatted_pname())
		now = datetime.datetime.now(datetime.timezone.utc)
		kdc_req_body = {}
		kdc_req_body['kdc-options'] = KDCOptions(set(['forwardable','renewable','renewable_ok', 'canonicalize']))
		kdc_req_body['realm'] = spn_user.domain.upper()
		kdc_req_body['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': spn_user.get_principalname()})
		kdc_req_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body['nonce'] = secrets.randbits(31)
		if override_etype:
			kdc_req_body['etype'] = override_etype
		else:
			if self.kerberos_cipher_type == -128:
				# we dunno how to do GSS api calls with -128 etype,
				# but we can request etype 23 here for which all is implemented
				kdc_req_body['etype'] = [23]
			else:
				kdc_req_body['etype'] = [self.kerberos_cipher_type]

		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno
		authenticator_data['crealm'] = Realm(self.kerberos_TGT['crealm'])
		authenticator_data['cname'] = self.kerberos_TGT['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)
		
		if is_linux:
			ac = AuthenticatorChecksum()
			ac.flags = 0
			ac.channel_binding = b'\x00'*16
			
			chksum = {}
			chksum['cksumtype'] = 0x8003
			chksum['checksum'] = ac.to_bytes()


			authenticator_data['cksum'] = Checksum(chksum)
			authenticator_data['seq-number'] = 0
		
		authenticator_data_enc = self.kerberos_cipher.encrypt(self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None)
		
		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ap-options'] = APOptions(set())
		ap_req['ticket'] = Ticket(self.kerberos_TGT['ticket'])
		ap_req['authenticator'] = EncryptedData({'etype': self.kerberos_cipher_type, 'cipher': authenticator_data_enc})
		
		pa_data_1 = {}
		pa_data_1['padata-type'] = PaDataType.TGS_REQ.value
		pa_data_1['padata-value'] = AP_REQ(ap_req).dump()
		
		
		kdc_req = {}
		kdc_req['pvno'] = krb5_pvno
		kdc_req['msg-type'] = MESSAGE_TYPE.KRB_TGS_REQ.value
		kdc_req['padata'] = [pa_data_1]
		kdc_req['req-body'] = KDC_REQ_BODY(kdc_req_body)
		
		req = TGS_REQ(kdc_req)
		logger.debug('Constructing TGS request to server')
		rep = self.ksoc.sendrecv(req.dump())
		if rep.name == 'KRB_ERROR':
			raise KerberosError(rep, 'get_TGS failed!')
		logger.debug('Got TGS reply, decrypting...')
		tgs = rep.native
		
		encTGSRepPart = EncTGSRepPart.load(self.kerberos_cipher.decrypt(self.kerberos_session_key, 8, tgs['enc-part']['cipher'])).native
		key = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'])
		
		self.ccache.add_tgs(tgs, encTGSRepPart)
		logger.debug('Got valid TGS reply')
		self.kerberos_TGS = tgs
		return tgs, encTGSRepPart, key

	def U2U(self, kdcopts = ['forwardable','renewable','canonicalize', 'enc-tkt-in-skey'], supp_enc_methods = [EncryptionType.DES_CBC_CRC,EncryptionType.DES_CBC_MD4,EncryptionType.DES_CBC_MD5,EncryptionType.DES3_CBC_SHA1,EncryptionType.ARCFOUR_HMAC_MD5,EncryptionType.AES256_CTS_HMAC_SHA1_96,EncryptionType.AES128_CTS_HMAC_SHA1_96]):
		if not self.kerberos_TGT:
			logger.debug('[U2U] TGT is not available! Fetching TGT...')
			self.get_TGT()

		supp_enc = self.credential.get_preferred_enctype(supp_enc_methods)
		now = datetime.datetime.now(datetime.timezone.utc)
		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno
		authenticator_data['crealm'] = Realm(self.kerberos_TGT['crealm'])
		authenticator_data['cname'] = self.kerberos_TGT['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)


		authenticator_data_enc = self.kerberos_cipher.encrypt(self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None)

		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ap-options'] = APOptions(set())
		ap_req['ticket'] = Ticket(self.kerberos_TGT['ticket'])
		ap_req['authenticator'] = EncryptedData({'etype': self.kerberos_cipher_type, 'cipher': authenticator_data_enc})

		pa_data_auth = {}
		pa_data_auth['padata-type'] = PaDataType.TGS_REQ.value
		pa_data_auth['padata-value'] = AP_REQ(ap_req).dump()

		
		krb_tgs_body = {}
		krb_tgs_body['kdc-options'] = KDCOptions(set(kdcopts))
		krb_tgs_body['sname'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': self.credential.username.split('/')})
		krb_tgs_body['realm'] = self.credential.domain.upper()
		krb_tgs_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		krb_tgs_body['nonce'] = secrets.randbits(31)
		krb_tgs_body['etype'] = [23] # dunno why it must be 23?
		krb_tgs_body['additional-tickets'] = [Ticket(self.kerberos_TGT['ticket'])]
		
		
		krb_tgs_req = {}
		krb_tgs_req['pvno'] = krb5_pvno
		krb_tgs_req['msg-type'] = MESSAGE_TYPE.KRB_TGS_REQ.value
		krb_tgs_req['padata'] = [pa_data_auth] #pa_for_user
		krb_tgs_req['req-body'] = KDC_REQ_BODY(krb_tgs_body)
		
		
		
		req = TGS_REQ(krb_tgs_req)
		logger.debug('[U2U] Sending request to server')
		
		reply = self.ksoc.sendrecv(req.dump())
		if reply.name == 'KRB_ERROR':
			emsg = '[U2U] failed!'
			if reply.native['error-code'] == 16:
				emsg = '[U2U] Failed to get U2U! Error code (16) indicates that delegation is not enabled for this account!'			
			raise KerberosError(reply, emsg)
		
		logger.debug('[U2U] Got reply, decrypting...')
		tgs = reply.native

		cipher = _enctype_table[int(tgs['ticket']['enc-part']['etype'])]
		encticket = tgs['ticket']['enc-part']['cipher']
		decdata = cipher.decrypt(self.kerberos_session_key, 2, encticket)
		decticket = EncTicketPart.load(decdata).native

		encTGSRepPart = EncTGSRepPart.load(self.kerberos_cipher.decrypt(self.kerberos_session_key, 8, tgs['enc-part']['cipher'])).native
		key = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'])
		self.ccache.add_tgs(tgs, encTGSRepPart)
		logger.debug('[U2U] Got valid TGS reply')

		return tgs, encTGSRepPart, key, decticket

	
	#https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/6a8dfc0c-2d32-478a-929f-5f9b1b18a169
	def S4U2self(self, user_to_impersonate, spn_user = None, kdcopts = ['forwardable','renewable','canonicalize'], supp_enc_methods = [EncryptionType.ARCFOUR_HMAC_MD5,EncryptionType.AES256_CTS_HMAC_SHA1_96,EncryptionType.AES128_CTS_HMAC_SHA1_96]):
		"""
		user_to_impersonate : KerberosTarget class
		"""		
		if not self.kerberos_TGT:
			logger.debug('[S4U2self] TGT is not available! Fetching TGT...')
			self.get_TGT()
		
		supp_enc = self.credential.get_preferred_enctype(supp_enc_methods)
		auth_package_name = 'Kerberos'
		now = datetime.datetime.now(datetime.timezone.utc)
		
		
		###### Calculating authenticator data
		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno
		authenticator_data['crealm'] = Realm(self.kerberos_TGT['crealm'])
		authenticator_data['cname'] = self.kerberos_TGT['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)
		
		authenticator_data_enc = self.kerberos_cipher.encrypt(self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None)
		
		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ap-options'] = APOptions(set())
		ap_req['ticket'] = Ticket(self.kerberos_TGT['ticket'])
		ap_req['authenticator'] = EncryptedData({'etype': self.kerberos_cipher_type, 'cipher': authenticator_data_enc})
		
		
		pa_data_auth = {}
		pa_data_auth['padata-type'] = PaDataType.TGS_REQ.value
		pa_data_auth['padata-value'] = AP_REQ(ap_req).dump()
		
		###### Calculating checksum data
		
		S4UByteArray = NAME_TYPE.PRINCIPAL.value.to_bytes(4, 'little', signed = False)
		S4UByteArray += user_to_impersonate.username.encode()
		S4UByteArray += user_to_impersonate.domain.encode()
		S4UByteArray += auth_package_name.encode()
		logger.debug('[S4U2self] S4UByteArray: %s' % S4UByteArray.hex())
		logger.debug('[S4U2self] S4UByteArray: %s' % S4UByteArray)
		
		chksum_data = _HMACMD5.checksum(self.kerberos_session_key, 17, S4UByteArray)
		logger.debug('[S4U2self] chksum_data: %s' % chksum_data.hex())
		
		
		chksum = {}
		chksum['cksumtype'] = int(CKSUMTYPE('HMAC_MD5'))
		chksum['checksum'] = chksum_data

		
		###### Filling out PA-FOR-USER data for impersonation
		pa_for_user_enc = {}
		pa_for_user_enc['userName'] = PrincipalName({'name-type': NAME_TYPE.PRINCIPAL.value, 'name-string': user_to_impersonate.get_principalname()})
		pa_for_user_enc['userRealm'] = user_to_impersonate.domain
		pa_for_user_enc['cksum'] = Checksum(chksum)
		pa_for_user_enc['auth-package'] = auth_package_name
		
		pa_for_user = {}
		pa_for_user['padata-type'] = int(PADATA_TYPE('PA-FOR-USER'))
		pa_for_user['padata-value'] = PA_FOR_USER_ENC(pa_for_user_enc).dump()
	
		###### Constructing body
		if spn_user is not None:
			if isinstance(spn_user, str):
				spn_user = [spn_user]
			elif isinstance(spn_user, list):
				spn_user = spn_user
			else:
				spn_user = spn_user.get_principalname()
		else:
			spn_user = self.credential.username.split('/')

		
		krb_tgs_body = {}
		krb_tgs_body['kdc-options'] = KDCOptions(set(kdcopts))
		krb_tgs_body['sname'] = PrincipalName({'name-type': NAME_TYPE.UNKNOWN.value, 'name-string': spn_user})
		krb_tgs_body['realm'] = self.credential.domain.upper()
		krb_tgs_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		krb_tgs_body['nonce'] = secrets.randbits(31)
		krb_tgs_body['etype'] = [self.kerberos_session_key.enctype] #[supp_enc.value] #selecting according to server's preferences
		
		
		krb_tgs_req = {}
		krb_tgs_req['pvno'] = krb5_pvno
		krb_tgs_req['msg-type'] = MESSAGE_TYPE.KRB_TGS_REQ.value
		krb_tgs_req['padata'] = [pa_data_auth, pa_for_user]
		krb_tgs_req['req-body'] = KDC_REQ_BODY(krb_tgs_body)
		
		req = TGS_REQ(krb_tgs_req)
		
		logger.debug('[S4U2self] Sending request to server')
		
		reply = self.ksoc.sendrecv(req.dump())
		if reply.name == 'KRB_ERROR':
			emsg = 'S4U2self failed!'
			if reply.native['error-code'] == 16:
				emsg = 'S4U2self: Failed to get S4U2self! Error code (16) indicates that delegation is not enabled for this account!'			
			raise KerberosError(reply, emsg)
		
		logger.debug('[S4U2self] Got reply, decrypting...')
		tgs = reply.native
		
		encTGSRepPart = EncTGSRepPart.load(self.kerberos_cipher.decrypt(self.kerberos_session_key, 8, tgs['enc-part']['cipher'])).native
		key = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'])
		
		self.ccache.add_tgs(tgs, encTGSRepPart)
		logger.debug('[S4U2self] Got valid TGS reply')
		self.kerberos_TGS = tgs
		return tgs, encTGSRepPart, key
				
	# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-sfu/c920c148-8a9c-42e9-b8e9-db5755cd281b
	def S4U2proxy(self, s4uself_ticket, spn_user, supp_enc_methods = [EncryptionType.ARCFOUR_HMAC_MD5,EncryptionType.AES256_CTS_HMAC_SHA1_96,EncryptionType.AES128_CTS_HMAC_SHA1_96]):
		logger.debug('[S4U2proxy] Impersonating %s' % '/'.join(spn_user.get_principalname()))
		now = datetime.datetime.now(datetime.timezone.utc)
		supp_enc = self.credential.get_preferred_enctype(supp_enc_methods)
		
		pa_pac_opts = {}
		pa_pac_opts['padata-type'] = int(PADATA_TYPE('PA-PAC-OPTIONS'))
		pa_pac_opts['padata-value'] = PA_PAC_OPTIONS({'value' : PA_PAC_OPTIONSTypes(set(['resource-based constrained delegation']))}).dump()

		
		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno
		authenticator_data['crealm'] = Realm(self.kerberos_TGT['crealm'])
		authenticator_data['cname'] = self.kerberos_TGT['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)
		
		authenticator_data_enc = self.kerberos_cipher.encrypt(self.kerberos_session_key, 7, Authenticator(authenticator_data).dump(), None)
		
		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ap-options'] = APOptions(set())
		ap_req['ticket'] = Ticket(self.kerberos_TGT['ticket'])
		ap_req['authenticator'] = EncryptedData({'etype': self.kerberos_cipher_type, 'cipher': authenticator_data_enc})
		
		pa_tgs_req = {}
		pa_tgs_req['padata-type'] = PaDataType.TGS_REQ.value
		pa_tgs_req['padata-value'] = AP_REQ(ap_req).dump()
		
		
		krb_tgs_body = {}
		#krb_tgs_body['kdc-options'] = KDCOptions(set(['forwardable','forwarded','renewable','renewable-ok', 'canonicalize']))
		krb_tgs_body['kdc-options'] = KDCOptions(set(['forwardable','renewable','constrained-delegation', 'canonicalize']))
		krb_tgs_body['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string': spn_user.get_principalname()})
		krb_tgs_body['realm'] = self.credential.domain.upper()
		krb_tgs_body['till'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		krb_tgs_body['nonce'] = secrets.randbits(31)
		krb_tgs_body['etype'] = [self.kerberos_session_key.enctype] #[supp_enc.value] #selecting according to server's preferences
		krb_tgs_body['additional-tickets'] = [s4uself_ticket]
		
		
		krb_tgs_req = {}
		krb_tgs_req['pvno'] = krb5_pvno
		krb_tgs_req['msg-type'] = MESSAGE_TYPE.KRB_TGS_REQ.value
		krb_tgs_req['padata'] = [pa_tgs_req, pa_pac_opts]
		krb_tgs_req['req-body'] = KDC_REQ_BODY(krb_tgs_body)
		
		req = TGS_REQ(krb_tgs_req)
		
		reply = self.ksoc.sendrecv(req.dump())
		if reply.name == 'KRB_ERROR':
			emsg = 'S4U2proxy failed!'
			if reply.native['error-code'] == 16:
				emsg = 'S4U2proxy: Failed to get S4U2proxy! Error code (16) indicates that delegation is not enabled for this account!'
			
			raise KerberosError(reply, emsg)
		
		logger.debug('[S4U2proxy] Got server reply, decrypting...')
		tgs = reply.native
		
		encTGSRepPart = EncTGSRepPart.load(self.kerberos_cipher.decrypt(self.kerberos_session_key, 8, tgs['enc-part']['cipher'])).native
		key = Key(encTGSRepPart['key']['keytype'], encTGSRepPart['key']['keyvalue'])
		
		self.ccache.add_tgs(tgs, encTGSRepPart)
		logger.debug('[S4U2proxy] Got valid TGS reply')

		return tgs, encTGSRepPart, key
		

	def construct_apreq(self, tgs, encTGSRepPart, sessionkey, flags = None, seq_number = 0, ap_opts = [], cb_data = None):
		return construct_apreq_from_tgs_tgt(
			tgs, 
			sessionkey, 
			self.kerberos_TGT, 
			flags = flags, 
			seq_number = seq_number, 
			ap_opts = ap_opts, 
			cb_data = cb_data
		)

	def getST(self, target_user, service_spn):
		tgs, encTGSRepPart, key = self.S4U2self(target_user)
		return self.S4U2proxy(tgs['ticket'], service_spn)

	def decrypt_asrep_cert(self, as_rep):
		
		def truncate_key(value, keysize):
			output = b''
			currentNum = 0
			while len(output) < keysize:
				currentDigest = hashlib.sha1(bytes([currentNum]) + value).digest()
				if len(output) + len(currentDigest) > keysize:
					output += currentDigest[:keysize - len(output)]
					break
				output += currentDigest
				currentNum += 1
			
			return output

		for pa in as_rep['padata']:
			if pa['padata-type'] == 17:
				pkasrep = PA_PK_AS_REP.load(pa['padata-value']).native
				break
		else:
			raise Exception('PA_PK_AS_REP not found!')

		try:
			sd = cms.SignedData.load(pkasrep['dhSignedData']).native
		except:
			sd = cms.SignedData.load(pkasrep['dhSignedData'][19:]).native # !!!!!!!!!!!!! TODO: CHECKTHIS!!! Sometimes there is an OID before the struct?!
	
		keyinfo = sd['encap_content_info']
		if keyinfo['content_type'] != '1.3.6.1.5.2.3.2':
			raise Exception('Keyinfo content type unexpected value')
		authdata = KDCDHKeyInfo.load(keyinfo['content']).native
		pubkey = int(''.join(['1'] + [str(x) for x in authdata['subjectPublicKey']]), 2)		

		pubkey = int.from_bytes(core.BitString(authdata['subjectPublicKey']).dump()[7:], 'big', signed = False) # !!!!!!!!!!!!! TODO: CHECKTHIS!!!
		shared_key = self.credential.dhparams.exchange(pubkey)
		
		server_nonce = pkasrep['serverDHNonce']
		fullKey = shared_key + self.credential.dhparams.dh_nonce + server_nonce

		etype = as_rep['enc-part']['etype']
		cipher = _enctype_table[etype]
		if etype == Enctype.AES256:
			self.pkinit_tkey = truncate_key(fullKey, 32)
		elif etype == Enctype.AES128:
			self.pkinit_tkey = truncate_key(fullKey, 16)
		elif etype == Enctype.RC4:
			raise NotImplementedError('RC4 key truncation documentation missing. it is different from AES')
			#self.pkinit_tkey = truncate_key(fullKey, 16)
		

		key = Key(cipher.enctype, self.pkinit_tkey)
		enc_data = as_rep['enc-part']['cipher']
		dec_data = cipher.decrypt(key, 3, enc_data)
		encasrep = EncASRepPart.load(dec_data).native
		cipher = _enctype_table[ int(encasrep['key']['keytype'])]
		session_key = Key(cipher.enctype, encasrep['key']['keyvalue'])
		return encasrep, session_key, cipher


	



from minikerberos.protocol.encryption import Key, Enctype, encrypt, decrypt, verify_checksum, Cksumtype, string_to_key, cf2, _zeropad, prf
from binascii import unhexlify as h

def test_encdec_aes():
	# AES128 encrypt and decrypt
	kb = h('9062430C8CDA3388922E6D6A509F5B7A')
	conf = h('94B491F481485B9A0678CD3C4EA386AD')
	keyusage = 2
	plain = '9 bytesss'.encode()
	ctxt = h('68FB9679601F45C78857B2BF820FD6E53ECA8D42FD4B1D7024A09205ABB7CD2E'
			 'C26C355D2F')
	k = Key(Enctype.AES128, kb)
	assert(encrypt(k, keyusage, plain, conf) == ctxt)
	assert(decrypt(k, keyusage, ctxt) == plain)

def test_encdec_aes256():
	# AES256 encrypt and decrypt
	kb = h('F1C795E9248A09338D82C3F8D5B567040B0110736845041347235B1404231398')
	conf = h('E45CA518B42E266AD98E165E706FFB60')
	keyusage = 4
	plain = '30 bytes bytes bytes bytes byt'.encode()
	ctxt = h('D1137A4D634CFECE924DBC3BF6790648BD5CFF7DE0E7B99460211D0DAEF3D79A'
			 '295C688858F3B34B9CBD6EEBAE81DAF6B734D4D498B6714F1C1D')
	k = Key(Enctype.AES256, kb)
	assert(encrypt(k, keyusage, plain, conf) == ctxt)
	assert(decrypt(k, keyusage, ctxt) == plain)

def test_aes128_checksum():
	# AES128 checksum
	kb = h('9062430C8CDA3388922E6D6A509F5B7A')
	keyusage = 3
	plain = 'eight nine ten eleven twelve thirteen'.encode()
	cksum = h('01A4B088D45628F6946614E3')
	k = Key(Enctype.AES128, kb)
	verify_checksum(Cksumtype.SHA1_AES128, k, keyusage, plain, cksum)

def test_aes256_checksum():
	# AES256 checksum
	kb = h('B1AE4CD8462AFF1677053CC9279AAC30B796FB81CE21474DD3DDBCFEA4EC76D7')
	keyusage = 4
	plain = 'fourteen'.encode()
	cksum = h('E08739E3279E2903EC8E3836')
	k = Key(Enctype.AES256, kb)
	verify_checksum(Cksumtype.SHA1_AES256, k, keyusage, plain, cksum)

def test_aes128_string_to_key():
	# AES128 string-to-key
	string = 'password'.encode()
	salt = 'ATHENA.MIT.EDUraeburn'.encode()
	params = h('00000002')
	kb = h('C651BF29E2300AC27FA469D693BDDA13')
	k = string_to_key(Enctype.AES128, string, salt, params)
	assert(k.contents == kb)

def test_aes256_string_to_key():
	# AES256 string-to-key
	string = b'X' * 64
	salt = 'pass phrase equals block size'.encode()
	params = h('000004B0')
	kb = h('89ADEE3608DB8BC71F1BFBFE459486B05618B70CBAE22092534E56C553BA4B34')
	k = string_to_key(Enctype.AES256, string, salt, params)
	assert(k.contents == kb)

def test_aes128_prf():
	# AES128 prf
	kb = h('77B39A37A868920F2A51F9DD150C5717')
	k = string_to_key(Enctype.AES128, 'key1'.encode(), 'key1'.encode())
	assert(prf(k, b'\x01\x61') == kb)

def test_aes256_prf():
	# AES256 prf
	kb = h('0D674DD0F9A6806525A4D92E828BD15A')
	k = string_to_key(Enctype.AES256, 'key2'.encode(), 'key2'.encode())
	assert(prf(k, b'\x02\x62') == kb)

def test_aes128_cf2():
	# AES128 cf2
	kb = h('97DF97E4B798B29EB31ED7280287A92A')
	k1 = string_to_key(Enctype.AES128, 'key1'.encode(), 'key1'.encode())
	k2 = string_to_key(Enctype.AES128, 'key2'.encode(), 'key2'.encode())
	k = cf2(Enctype.AES128, k1, k2, b'a', b'b')
	assert(k.contents == kb)

def test_aes256_cf2():
	# AES256 cf2
	kb = h('4D6CA4E629785C1F01BAF55E2E548566B9617AE3A96868C337CB93B5E72B1C7B')
	k1 = string_to_key(Enctype.AES256, 'key1'.encode(), 'key1'.encode())
	k2 = string_to_key(Enctype.AES256, 'key2'.encode(), 'key2'.encode())
	k = cf2(Enctype.AES256, k1, k2, b'a', b'b')
	assert(k.contents == kb)

def test_des3_encrypt_decrypt():
	# DES3 encrypt and decrypt
	kb = h('0DD52094E0F41CECCB5BE510A764B35176E3981332F1E598')
	conf = h('94690A17B2DA3C9B')
	keyusage = 3
	plain = b'13 bytes byte'
	ctxt = h('839A17081ECBAFBCDC91B88C6955DD3C4514023CF177B77BF0D0177A16F705E8'
			 '49CB7781D76A316B193F8D30')
	k = Key(Enctype.DES3, kb)
	assert(encrypt(k, keyusage, plain, conf) == ctxt)
	assert(decrypt(k, keyusage, ctxt) == _zeropad(plain, 8))

def test_des3_string_to_key():
	# DES3 string-to-key
	string = 'password'.encode()
	salt = 'ATHENA.MIT.EDUraeburn'.encode()
	kb = h('850BB51358548CD05E86768C313E3BFEF7511937DCF72C3E')
	k = string_to_key(Enctype.DES3, string, salt)
	assert(k.contents == kb)

def test_des3_checksum():
	# DES3 checksum
	kb = h('7A25DF8992296DCEDA0E135BC4046E2375B3C14C98FBC162')
	keyusage = 2
	plain = 'six seven'.encode()
	cksum = h('0EEFC9C3E049AABC1BA5C401677D9AB699082BB4')
	k = Key(Enctype.DES3, kb)
	verify_checksum(Cksumtype.SHA1_DES3, k, keyusage, plain, cksum)

def test_des3_cf2():
	# DES3 cf2
	kb = h('E58F9EB643862C13AD38E529313462A7F73E62834FE54A01')
	k1 = string_to_key(Enctype.DES3, 'key1'.encode(), 'key1'.encode())
	k2 = string_to_key(Enctype.DES3, 'key2'.encode(), 'key2'.encode())
	k = cf2(Enctype.DES3, k1, k2, b'a', b'b')
	assert(k.contents == kb)

def test_rc4_encrypt_decrypt():
	# RC4 encrypt and decrypt
	kb = h('68F263DB3FCE15D031C9EAB02D67107A')
	conf = h('37245E73A45FBF72')
	keyusage = 4
	plain = b'30 bytes bytes bytes bytes byt'
	ctxt = h('95F9047C3AD75891C2E9B04B16566DC8B6EB9CE4231AFB2542EF87A7B5A0F260'
			 'A99F0460508DE0CECC632D07C354124E46C5D2234EB8')
	k = Key(Enctype.RC4, kb)
	assert(encrypt(k, keyusage, plain, conf) == ctxt)
	assert(decrypt(k, keyusage, ctxt) == plain)

def test_rc4_string_to_key():
	# RC4 string-to-key
	string = 'foo'.encode()
	kb = h('AC8E657F83DF82BEEA5D43BDAF7800CC')
	k = string_to_key(Enctype.RC4, string, None)
	assert(k.contents == kb)

def test_rc4_checksum():
	# RC4 checksum
	kb = h('F7D3A155AF5E238A0B7A871A96BA2AB2')
	keyusage = 6
	plain = 'seventeen eighteen nineteen twenty'.encode()
	cksum = h('EB38CC97E2230F59DA4117DC5859D7EC')
	k = Key(Enctype.RC4, kb)
	verify_checksum(Cksumtype.HMAC_MD5, k, keyusage, plain, cksum)

def test_rc4_cf2():
	# RC4 cf2
	kb = h('24D7F6B6BAE4E5C00D2082C5EBAB3672')
	k1 = string_to_key(Enctype.RC4, 'key1'.encode(), 'key1'.encode())
	k2 = string_to_key(Enctype.RC4, 'key2'.encode(), 'key2'.encode())
	k = cf2(Enctype.RC4, k1, k2, b'a', b'b')
	assert(k.contents == kb)

def test_des_string_to_key():
	# DES string-to-key
	string = 'password'.encode()
	salt = 'ATHENA.MIT.EDUraeburn'.encode()
	kb = h('cbc22fae235298e3')
	k = string_to_key(Enctype.DES_MD5, string, salt)
	assert(k.contents == kb)

def test_des_string_to_key():
	# DES string-to-key
	string = 'potatoe'.encode()
	salt = 'WHITEHOUSE.GOVdanny'.encode()
	kb = h('df3d32a74fd92a01')
	k = string_to_key(Enctype.DES_MD5, string, salt)
	assert(k.contents == kb)
	print('all tests passed!')
	
	



import os
import io

from asn1crypto import core
from minikerberos.protocol.asn1_structs import EncryptionKey, Checksum, KerberosTime, Realm

TAG = 'explicit'

# class
UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2


########
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-kile/1aeca7fb-d6b4-4402-8fa4-6ec3e955c16e
class KERB_AD_RESTRICTION_ENTRY(core.Sequence):
    _fields = [
        ('restriction-type', core.Integer, {'tag_type': TAG, 'tag': 0}),
        ('restriction', core.OctetString, {'tag_type': TAG, 'tag': 1}),
    ]
    
class KERB_AD_RESTRICTION_ENTRYS(core.SequenceOf):
    _child_spec = KERB_AD_RESTRICTION_ENTRY


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-kile/25fabd02-560d-4c1f-8f42-b32e9d97996a
class KERB_ERROR_DATA(core.Sequence):
    _fields = [
        ('data-type', core.Integer, {'tag_type': TAG, 'tag': 1}),
        ('data-value', core.OctetString, {'tag_type': TAG, 'tag': 2, 'optional': True}),
    ]

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-kile/765795ba-9e05-4220-9bd3-b34464e413a7
class KERB_PA_PAC_REQUEST(core.Sequence):
    _fields = [
        ('include-pac', core.Boolean, {'tag_type': TAG, 'tag': 0}),
    ]


## implementation specific??????
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-kile/2a01b297-c47f-4547-9268-cf589aedd063
#class KERB_LOCAL(core.Sequence):
#    _fields = [
#    ]


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-kile/ec551137-c5e5-476a-9c89-e0029473c41b
class LSAP_TOKEN_INFO_INTEGRITY:
	def __init__(self, machine_id = os.urandom(32)):
		self.Flags = None # unsigned long
		self.TokenIL = None # unsigned long
		self.MachineID = machine_id # KILE implements a 32-byte binary random string machine ID.

	def to_bytes(self):
		t = self.Flags.to_bytes(4, byteorder='little', signed = False)
		t += self.TokenIL.to_bytes(4, byteorder='little', signed = False)
		t += self.MachineID
		return t
	
	@staticmethod
	def from_bytes(data):
		return LSAP_TOKEN_INFO_INTEGRITY.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		msg = LSAP_TOKEN_INFO_INTEGRITY()
		msg.Flags = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.TokenIL = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.MachineID = buff.read(32)
		return msg


class KERB_KEY_LIST_REP(core.SequenceOf):
	_child_spec = EncryptionKey

class KERB_KEY_LIST_REQ(core.SequenceOf):
	_child_spec = core.Integer



#### TODO
"""
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-kile/15fa77fb-deaa-487d-b685-1310afe45ca1
 typedef struct KERB_EXT_ERROR {
     unsigned long status;
     unsigned long reserved;
     unsigned long flags;
 } KERB_EXT_ERROR;

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-kile/25fabd02-560d-4c1f-8f42-b32e9d97996a



"""


import logging
import asyncio
from minikerberos.common.factory import KerberosClientFactory, kerberos_url_help_epilog
from minikerberos.common.kirbi import Kirbi

async def getTGT(kerberos_url, kirbifile = None, ccachefile = None, nopac = False):
	cu = KerberosClientFactory.from_url(kerberos_url)
	client = cu.get_client()
	logging.debug('Getting TGT')
	
	await client.get_TGT(with_pac=nopac)
	if ccachefile is not None:
		client.ccache.to_file(ccachefile)
		print('TGT stored in ccache file %s' % ccachefile)
	
	kirbi = Kirbi.from_ticketdata(client.kerberos_TGT, client.kerberos_TGT_encpart)
	print(str(kirbi))
	if kirbifile is not None:
		kirbi.to_file(kirbifile)
		
	logging.info('Done!')

def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='Polls the kerberos service for a TGT for the sepcified user', formatter_class=argparse.RawDescriptionHelpFormatter, epilog = kerberos_url_help_epilog)
	parser.add_argument('--nopac', action='store_false', help="Don't request a PAC in the TGT")
	parser.add_argument('--ccache', help='CCACHE file to store the TGT ticket in, otherwise kirbi will be printed to stdout')
	parser.add_argument('--kirbi', help='kirbi file to store the TGT ticket in, otherwise kirbi will be printed to stdout')
	parser.add_argument('kerberos_url', help='the kerberos target string. ')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	
	args = parser.parse_args()
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig(level=1)
	
	asyncio.run(getTGT(args.kerberos_url, args.kirbi, args.ccache, args.nopac))
	
	
if __name__ == '__main__':
	main()



# Kudos:
# Parts of this code was inspired by the following project by @rubin_mor
# https://github.com/morRubin/AzureADJoinedMachinePTC
# 

# TODO: code currently supports RSA+DH+SHA1 , add support for other mechanisms


import os
import datetime
import secrets
import platform

from unicrypto import hashlib
from asn1crypto import cms
from asn1crypto import algos
from asn1crypto import core
from asn1crypto import x509
from asn1crypto import keys

from oscrypto.keys import parse_pkcs12
from oscrypto.asymmetric import rsa_pkcs1v15_sign, load_private_key

from minikerberos.protocol.constants import NAME_TYPE, MESSAGE_TYPE, PaDataType
from minikerberos.protocol.encryption import Enctype, _checksum_table, _enctype_table, Key
from minikerberos.protocol.structures import AuthenticatorChecksum
from minikerberos.protocol.asn1_structs import KDC_REQ_BODY, PrincipalName, HostAddress, \
	KDCOptions, EncASRepPart, AP_REQ, AuthorizationData, Checksum, krb5_pvno, Realm, \
	EncryptionKey, Authenticator, Ticket, APOptions, EncryptedData, AS_REQ, AP_REP
from minikerberos.protocol.rfc4556 import PKAuthenticator, AuthPack, Dunno1, Dunno2, MetaData, Info, CertIssuer, CertIssuers, PA_PK_AS_REP, KDCDHKeyInfo
from minikerberos.protocol.rfc_iakerb import KRB_FINISHED
from minikerberos.protocol.mskile import LSAP_TOKEN_INFO_INTEGRITY, KERB_AD_RESTRICTION_ENTRY, KERB_AD_RESTRICTION_ENTRYS
from minikerberos.gssapi.gssapi import GSSAPIFlags

def length_encode(x):
	if x <= 127:
		return x.to_bytes(1, 'big', signed = False)
	else:
		lb = x.to_bytes((x.bit_length() + 7) // 8, 'big')
		t = (0x80 | len(lb)).to_bytes(1, 'big', signed = False)
		return t+lb


class DirtyDH:
	def __init__(self):
		self.p = None
		self.g = None
		self.shared_key = None
		self.shared_key_int = None
		self.private_key = os.urandom(32)
		self.private_key_int = int(self.private_key.hex(), 16)
		self.dh_nonce = os.urandom(32)

	@staticmethod
	def from_params(p, g):
		dd = DirtyDH()
		dd.p = p
		dd.g = g
		return dd

	@staticmethod
	def from_dict(dhp):
		dd = DirtyDH()
		dd.p = dhp['p']
		dd.g = dhp['g']
		return dd

	@staticmethod
	def from_asn1(asn1_bytes):
		dhp = algos.DHParameters.load(asn1_bytes).native
		return DirtyDH.from_dict(dhp)
		
	
	def get_public_key(self):
		#y = g^x mod p
		return pow(self.g, self.private_key_int, self.p)
	
	def exchange(self, bob_int):
		self.shared_key_int = pow(bob_int, self.private_key_int, self.p)
		x = hex(self.shared_key_int)[2:]
		if len(x) % 2 != 0:
			x = '0' + x
		self.shared_key = bytes.fromhex(x)
		return self.shared_key


class PKINIT:
	def __init__(self):
		self.privkeyinfo = None
		self.certificate = None
		self.extra_certs = None
		self.user_sid = None
		self.user_name = None
		self.issuer = None
		self.cname = None
		self.diffie = None
		self.__hcert = None


	def init_windows_cert(self, username, certstore_name = 'MY', cert_serial = None):
		from minikerberos.common.windows.crypt32 import find_cert_by_cn
		self.certificate, self.__hcert = find_cert_by_cn(username, certstore_name)
		

	@staticmethod
	def from_windows_certstore(username, certstore_name = 'MY', cert_serial = None, dh_params = None):
		pkinit = PKINIT()
		pkinit.init_windows_cert(username, certstore_name = certstore_name, cert_serial = cert_serial)
		pkinit.setup(dh_params = dh_params)
		return pkinit


	@staticmethod
	def from_pfx(pfxfile, pfxpass, dh_params = None):
		pkinit = PKINIT()
		#print('Loading pfx12')		
		if isinstance(pfxpass, str):
			pfxpass = pfxpass.encode()
		with open(pfxfile, 'rb') as f:
			pkinit.privkeyinfo, pkinit.certificate, pkinit.extra_certs = parse_pkcs12(f.read(), password = pfxpass)
			pkinit.privkey = load_private_key(pkinit.privkeyinfo)
		#print('pfx12 loaded!')
		pkinit.setup(dh_params = dh_params)
		return pkinit
	
	def setup(self, dh_params = None):
		# parsing ceritficate to get basic info that will be needed to construct the asreq
		# this has two components
		for x in self.certificate.subject.native['common_name']:
			if x.startswith("S-1-12"):
				self.user_sid = x
			elif x.find('@') != -1:
				self.user_name = x

		self.issuer = self.certificate.issuer.native['common_name']
		self.cname = '\\'.join(['AzureAD', self.issuer, self.user_sid])

		#print('cert issuer: %s' % pkinit.issuer)
		#print('cert user_name: %s' % pkinit.user_name)
		#print('cert user_sid: %s' % pkinit.user_sid)
		#print('cert cname: %s' % pkinit.cname)

		if dh_params is None:
			print('Generating DH params...')
			self.diffie = DirtyDH.from_dict( generate_dh_parameters(1024).native)
			print('DH params generated.')
		else:
			#print('Loading default DH params...')
			if isinstance(dh_params, dict):
				self.diffie = DirtyDH.from_dict(dh_params)
			elif isinstance(dh_params, bytes):
				self.diffie = DirtyDH.from_asn1(dh_params)
			elif isinstance(dh_params, DirtyDH):
				self.diffie = dh_params
			else:
				raise Exception('DH params must be either a bytearray or a dict')


	def build_asreq(self, target = None, cname = None, kdcopts = ['forwardable','renewable','proxiable', 'canonicalize']):
		if isinstance(kdcopts, list):
			kdcopts = set(kdcopts)
		if cname is not None:
			if isinstance(cname, str):
				cname = [cname]
		else:
			cname = [self.cname]
		
		if target is not None:
			if isinstance(target, str):
				target = [target]
		else:
			target = ['127.0.0.1']

		now = datetime.datetime.now(datetime.timezone.utc)

		kdc_req_body_data = {}
		kdc_req_body_data['kdc-options'] = KDCOptions(kdcopts)
		kdc_req_body_data['cname'] = PrincipalName({'name-type': NAME_TYPE.MS_PRINCIPAL.value, 'name-string': cname})
		kdc_req_body_data['realm'] = 'WELLKNOWN:PKU2U'
		kdc_req_body_data['sname'] = PrincipalName({'name-type': NAME_TYPE.MS_PRINCIPAL.value, 'name-string': target})
		kdc_req_body_data['till']  = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body_data['rtime'] = (now + datetime.timedelta(days=1)).replace(microsecond=0)
		kdc_req_body_data['nonce'] = secrets.randbits(31)
		kdc_req_body_data['etype'] = [18,17] # 23 breaks...
		kdc_req_body_data['addresses'] = [HostAddress({'addr-type': 20, 'address': b'127.0.0.1'})] # not sure if this is needed
		kdc_req_body = KDC_REQ_BODY(kdc_req_body_data)


		checksum = hashlib.sha1(kdc_req_body.dump()).digest()
		
		authenticator = {}
		authenticator['cusec'] = now.microsecond
		authenticator['ctime'] = now.replace(microsecond=0)
		authenticator['nonce'] = secrets.randbits(31)
		authenticator['paChecksum'] = checksum
		

		dp = {}
		dp['p'] = self.diffie.p
		dp['g'] = self.diffie.g
		dp['q'] = 0 # mandatory parameter, but it is not needed

		pka = {}
		pka['algorithm'] = '1.2.840.10046.2.1'
		pka['parameters'] = keys.DomainParameters(dp)
		
		spki = {}
		spki['algorithm'] = keys.PublicKeyAlgorithm(pka)
		spki['public_key'] = self.diffie.get_public_key()

		
		authpack = {}
		authpack['pkAuthenticator'] = PKAuthenticator(authenticator)
		authpack['clientPublicValue'] = keys.PublicKeyInfo(spki)
		authpack['clientDHNonce'] = self.diffie.dh_nonce
		
		authpack = AuthPack(authpack)
		signed_authpack = self.sign_authpack(authpack.dump(), wrap_signed = False)
		
		# ??????? This is absolutely nonsense, 
		payload = length_encode(len(signed_authpack)) + signed_authpack
		payload = b'\x80' + payload
		signed_authpack = b'\x30' + length_encode(len(payload)) + payload
		
		pa_data_1 = {}
		pa_data_1['padata-type'] = PaDataType.PK_AS_REQ.value
		pa_data_1['padata-value'] = signed_authpack 

		asreq = {}
		asreq['pvno'] = 5
		asreq['msg-type'] = 10
		asreq['padata'] = [pa_data_1]
		asreq['req-body'] = kdc_req_body

		return AS_REQ(asreq).dump()	

	def build_apreq(self, asrep, session_key, cipher, subkey_data, krb_finished_data, flags = GSSAPIFlags.GSS_C_MUTUAL_FLAG | GSSAPIFlags.GSS_C_INTEG_FLAG  | GSSAPIFlags.GSS_C_EXTENDED_ERROR_FLAG):
		

		# TODO: https://www.ietf.org/rfc/rfc4757.txt
		#subkey_data = {}
		#subkey_data['keytype'] = Enctype.AES256
		#subkey_data['keyvalue'] = os.urandom(32)

		subkey_cipher = _enctype_table[subkey_data['keytype']]
		subkey_key = Key(subkey_cipher.enctype, subkey_data['keyvalue'])
		subkey_checksum = _checksum_table[16] # ChecksumTypes.hmac_sha1_96_aes256

		krb_finished_checksum_data = {}
		krb_finished_checksum_data['cksumtype'] = 16
		krb_finished_checksum_data['checksum'] = subkey_checksum.checksum(subkey_key, 41, krb_finished_data)

		krb_finished_data = {}
		krb_finished_data['gss-mic'] = Checksum(krb_finished_checksum_data)

		krb_finished = KRB_FINISHED(krb_finished_data).dump()

		a = 2
		extensions_data = a.to_bytes(4, byteorder='big', signed=True) + len(krb_finished).to_bytes(4, byteorder='big', signed=True) + krb_finished

		ac = AuthenticatorChecksum()
		ac.flags = flags
		ac.channel_binding = b'\x00'*16
		chksum = {}
		chksum['cksumtype'] = 0x8003
		chksum['checksum'] = ac.to_bytes() + extensions_data

		tii = LSAP_TOKEN_INFO_INTEGRITY()
		tii.Flags = 1
		tii.TokenIL = 0x00002000 # Medium integrity
		tii.MachineID = bytes.fromhex('7e303fffe6bff25146addca4fbddf1b94f1634178eb4528fb2731c669ca23cde')

		restriction_data = {}
		restriction_data['restriction-type'] = 0
		restriction_data['restriction'] = tii.to_bytes()
		restriction_data = KERB_AD_RESTRICTION_ENTRY(restriction_data)

		x = KERB_AD_RESTRICTION_ENTRYS([restriction_data]).dump()
		restrictions = AuthorizationData([{ 'ad-type' : 141, 'ad-data' : x}]).dump()

		

		now = datetime.datetime.now(datetime.timezone.utc)
		authenticator_data = {}
		authenticator_data['authenticator-vno'] = krb5_pvno 
		authenticator_data['crealm'] = Realm(asrep['crealm'])
		authenticator_data['cname'] = asrep['cname']
		authenticator_data['cusec'] = now.microsecond
		authenticator_data['ctime'] = now.replace(microsecond=0)
		authenticator_data['subkey'] = EncryptionKey(subkey_data)
		authenticator_data['seq-number'] = 682437742 #??? TODO: check this!
		authenticator_data['authorization-data'] = AuthorizationData([{'ad-type': 1, 'ad-data' : restrictions}])
		authenticator_data['cksum'] = Checksum(chksum)
		
		
		#print('Authenticator(authenticator_data).dump()')
		#print(Authenticator(authenticator_data).dump().hex())

		authenticator_data_enc = cipher.encrypt(session_key, 11, Authenticator(authenticator_data).dump(), None)
		
		ap_opts = ['mutual-required']

		ap_req = {}
		ap_req['pvno'] = krb5_pvno
		ap_req['msg-type'] = MESSAGE_TYPE.KRB_AP_REQ.value
		ap_req['ticket'] = Ticket(asrep['ticket'])
		ap_req['ap-options'] = APOptions(set(ap_opts))
		ap_req['authenticator'] = EncryptedData({'etype': session_key.enctype, 'cipher': authenticator_data_enc})
		
		#pprint('AP_REQ \r\n%s' % AP_REQ(ap_req).native)
		
		#print(AP_REQ(ap_req).dump().hex())
		#input()

		return AP_REQ(ap_req).dump()

	def sign_authpack(self, data, wrap_signed = False):
		if self.__hcert is not None:
			from minikerberos.common.windows.crypt32 import pkcs7_sign

			return pkcs7_sign(self.__hcert, data)
		return self.sign_authpack_native(self, data, wrap_signed)

	def sign_authpack_native(self, data, wrap_signed = False):
		"""
		Creating PKCS7 blob which contains the following things:

		1. 'data' blob which is an ASN1 encoded "AuthPack" structure
		2. the certificate used to sign the data blob
		3. the singed 'signed_attrs' structure (ASN1) which points to the "data" structure (in point 1)
		"""
		
		da = {}
		da['algorithm'] = algos.DigestAlgorithmId('1.3.14.3.2.26') # for sha1

		si = {}
		si['version'] = 'v1'
		si['sid'] = cms.IssuerAndSerialNumber({
			'issuer':  self.certificate.issuer,
			'serial_number':  self.certificate.serial_number,
		})


		si['digest_algorithm'] = algos.DigestAlgorithm(da)
		si['signed_attrs'] = [
			cms.CMSAttribute({'type': 'content_type', 'values': ['1.3.6.1.5.2.3.1']}), # indicates that the encap_content_info's authdata struct (marked with OID '1.3.6.1.5.2.3.1' is signed )
			cms.CMSAttribute({'type': 'message_digest', 'values': [hashlib.sha1(data).digest()]}), ### hash of the data, the data itself will not be signed, but this block of data will be.
		]
		si['signature_algorithm'] = algos.SignedDigestAlgorithm({'algorithm' : '1.2.840.113549.1.1.1'})
		si['signature'] = rsa_pkcs1v15_sign(self.privkey,  cms.CMSAttributes(si['signed_attrs']).dump(), "sha1")

		ec = {}
		ec['content_type'] = '1.3.6.1.5.2.3.1'
		ec['content'] = data

		sd = {}
		sd['version'] = 'v3'
		sd['digest_algorithms'] = [algos.DigestAlgorithm(da)] # must have only one
		sd['encap_content_info'] = cms.EncapsulatedContentInfo(ec)
		sd['certificates'] = [self.certificate]
		sd['signer_infos'] = cms.SignerInfos([cms.SignerInfo(si)])
		
		if wrap_signed is True:
			ci = {}
			ci['content_type'] = '1.2.840.113549.1.7.2' # signed data OID
			ci['content'] = cms.SignedData(sd)
			return cms.ContentInfo(ci).dump()

		return cms.SignedData(sd).dump()
		
	
	
	
	
	def decrypt_asrep(self, as_rep):
		

		
		def truncate_key(value, keysize):
			output = b''
			currentNum = 0
			while len(output) < keysize:
				currentDigest = hashlib.sha1(bytes([currentNum]) + value).digest()
				if len(output) + len(currentDigest) > keysize:
					output += currentDigest[:keysize - len(output)]
					break
				output += currentDigest
				currentNum += 1
			
			return output

		for pa in as_rep['padata']:
			if pa['padata-type'] == 17:
				pkasrep = PA_PK_AS_REP.load(pa['padata-value']).native
				break
		else:
			raise Exception('PA_PK_AS_REP not found!')

		sd = cms.SignedData.load(pkasrep['dhSignedData']).native
		keyinfo = sd['encap_content_info']
		if keyinfo['content_type'] != '1.3.6.1.5.2.3.2':
			raise Exception('Keyinfo content type unexpected value')
		authdata = KDCDHKeyInfo.load(keyinfo['content']).native
		pubkey = int(''.join(['1'] + [str(x) for x in authdata['subjectPublicKey']]), 2)		

		pubkey = int.from_bytes(core.BitString(authdata['subjectPublicKey']).dump()[7:], 'big', signed = False)
		shared_key = self.diffie.exchange(pubkey)
		
		server_nonce = pkasrep['serverDHNonce']
		fullKey = shared_key + self.diffie.dh_nonce + server_nonce

		etype = as_rep['enc-part']['etype']
		cipher = _enctype_table[etype]
		if etype == Enctype.AES256:
			t_key = truncate_key(fullKey, 32)
		elif etype == Enctype.AES128:
			t_key = truncate_key(fullKey, 16)
		elif etype == Enctype.RC4:
			raise NotImplementedError('RC4 key truncation documentation missing. it is different from AES')
			#t_key = truncate_key(fullKey, 16)
		

		key = Key(cipher.enctype, t_key)
		enc_data = as_rep['enc-part']['cipher']
		dec_data = cipher.decrypt(key, 3, enc_data)
		encasrep = EncASRepPart.load(dec_data).native
		cipher = _enctype_table[ int(encasrep['key']['keytype'])]
		session_key = Key(cipher.enctype, encasrep['key']['keyvalue'])
		return encasrep, session_key, cipher

	
	def get_metadata(self, target = None):
		


		if target is not None:
			if isinstance(target, str):
				target = [target]
		else:
			target = ['127.0.0.1']

		
		ci = {}
		ci['type'] = '2.5.4.3'
		ci['value'] = self.issuer

		a = Dunno1([ci])
		ci = Dunno2([a])

		info = {}
		info['pku2u'] = 'WELLKNOWN:PKU2U'
		info['clientInfo'] = PrincipalName({'name-type': NAME_TYPE.MS_PRINCIPAL.value, 'name-string': target})

		md = {}
		md['Info'] = Info(info)
		md['1'] = [CertIssuer({'data' : ci.dump()})]

		return MetaData(md).dump()


import tempfile
import pathlib
import asyncio
from minikerberos.common.ccache import CCACHE
from minikerberos.common.kirbi import Kirbi
import pytest
import shutil
import os
from .config import *

def test_ccacheroast():
	from minikerberos.examples.ccacheroast import ccacheroast
	ccache = CCACHE()
	with tempfile.NamedTemporaryFile() as f:
		for kirbifile in get_testfiles_kirbi():
			kirbi = Kirbi.from_file(kirbifile)
			ccache.add_kirbi(kirbi)
		
		ccache.to_file(f.name)
		ccacheroast(f.name)
		assert len(ccache.get_hashes()) > 0

# testcases for spnroast
@pytest.mark.asyncio
async def test_spnroast_base():
	from minikerberos.examples.spnroast import spnroast
	with tempfile.NamedTemporaryFile() as f:
		x = await spnroast(KERBEROS_CONN_URL_PW, KERBEROS_KERBEROAST_USER, KERBEROS_DOMAIN, f.name)
		f.flush()
		f.seek(0,0)
		data = f.read()
		assert len(data) > 0

@pytest.mark.asyncio
async def test_spnroast_23():
	from minikerberos.examples.spnroast import spnroast
	with tempfile.NamedTemporaryFile() as f:
		x = await spnroast(KERBEROS_CONN_URL_PW, KERBEROS_KERBEROAST_USER, KERBEROS_DOMAIN, f.name, etypes=23)
		f.flush()
		f.seek(0,0)
		data = f.read()
		assert len(data) > 0

@pytest.mark.asyncio
async def test_spnroast_all():
	from minikerberos.examples.spnroast import spnroast
	with tempfile.NamedTemporaryFile() as f:
		x = await spnroast(KERBEROS_CONN_URL_PW, KERBEROS_KERBEROAST_USER, KERBEROS_DOMAIN, f.name, etypes='23,17,18')
		f.flush()
		f.seek(0,0)
		data = f.read()
		assert len(data) > 0

@pytest.mark.asyncio
async def test_spnroast_list():
	from minikerberos.examples.spnroast import spnroast
	with tempfile.NamedTemporaryFile() as f:
		x = await spnroast(KERBEROS_CONN_URL_PW, [KERBEROS_KERBEROAST_USER], KERBEROS_DOMAIN, f.name)
		f.flush()
		f.seek(0,0)
		data = f.read()
		assert len(data) > 0

@pytest.mark.asyncio
async def test_spnroast_file():
	from minikerberos.examples.spnroast import spnroast
	with tempfile.NamedTemporaryFile('w') as d:
		d.write(KERBEROS_KERBEROAST_USER)
		d.flush()
		with tempfile.NamedTemporaryFile() as f:
			x = await spnroast(KERBEROS_CONN_URL_PW, d.name, KERBEROS_DOMAIN, f.name)
			f.flush()
			f.seek(0,0)
			data = f.read()
			assert len(data) > 0

@pytest.mark.asyncio
async def test_spnroast_filedomain():
	from minikerberos.examples.spnroast import spnroast
	with tempfile.NamedTemporaryFile('w') as d:
		d.write(KERBEROS_KERBEROAST_USER + '@' + KERBEROS_DOMAIN)
		d.flush()
		with tempfile.NamedTemporaryFile() as f:
			x = await spnroast(KERBEROS_CONN_URL_PW, d.name, None, f.name)
			f.flush()
			f.seek(0,0)
			data = f.read()
			assert len(data) > 0

@pytest.mark.asyncio
async def test_spnroast_listdomain():
	from minikerberos.examples.spnroast import spnroast
	with tempfile.NamedTemporaryFile() as f:
		x = await spnroast(KERBEROS_CONN_URL_PW, [KERBEROS_KERBEROAST_USER + '@' + KERBEROS_DOMAIN], KERBEROS_DOMAIN, f.name)
		f.flush()
		f.seek(0,0)
		data = f.read()
		assert len(data) > 0

def test_kirbi2ccache():
	from minikerberos.examples.kirbi2ccache import kirbi2ccache
	ccache = CCACHE()
	for kirbifile in get_testfiles_kirbi():
		kirbi = Kirbi.from_file(kirbifile)
		ccache.add_kirbi(kirbi)
	with tempfile.NamedTemporaryFile() as f:
		kirbi2ccache(kirbifile, f.name)
		ccache2 = CCACHE.from_file(f.name)
		assert len(ccache2.get_hashes()) > 0

@pytest.mark.asyncio
async def test_gettgt():
	from minikerberos.examples.getTGT import getTGT
	with tempfile.NamedTemporaryFile() as d:
		with tempfile.NamedTemporaryFile() as f:
			await getTGT(KERBEROS_CONN_URL_PW, d.name, f.name)
			ccache = CCACHE.from_file(f.name)
			assert len(ccache.get_hashes()) > 0
		kirbi = Kirbi.from_file(d.name)
		assert kirbi.kirbiobj is not None

@pytest.mark.asyncio
async def test_gettgt_nopac():
	from minikerberos.examples.getTGT import getTGT
	with tempfile.NamedTemporaryFile() as d:
		with tempfile.NamedTemporaryFile() as f:
			await getTGT(KERBEROS_CONN_URL_PW, d.name, f.name, nopac=True)
			ccache = CCACHE.from_file(f.name)
			assert len(ccache.get_hashes()) > 0
		kirbi = Kirbi.from_file(d.name)
		assert kirbi.kirbiobj is not None

@pytest.mark.asyncio
async def test_gettgs():
	from minikerberos.examples.getTGS import getTGS
	with tempfile.NamedTemporaryFile() as d:
		with tempfile.NamedTemporaryFile() as f:
			await getTGS(KERBEROS_CONN_URL_PW, KERBEROS_TGS_SPN ,d.name, f.name)
			ccache = CCACHE.from_file(f.name)
			assert len(ccache.get_hashes()) > 0
		kirbi = Kirbi.from_file(d.name)
		assert kirbi.kirbiobj is not None

@pytest.mark.asyncio
async def test_getS4U2Self():
	from minikerberos.examples.getS4U2self import getS4U2self
	with tempfile.NamedTemporaryFile() as d:
		with tempfile.NamedTemporaryFile() as f:
			await getS4U2self(KERBEROS_CONN_URL_MACHINE, KERBEROS_DELEGATION_SPN_SELF, KERBEROS_DELEGATION_USER_SELF,d.name, f.name)
			ccache = CCACHE.from_file(f.name)
			assert len(ccache.get_hashes()) > 0
		kirbi = Kirbi.from_file(d.name)
		assert kirbi.kirbiobj is not None

@pytest.mark.asyncio
async def test_getS4U2Self_ccache():
	import os
	from minikerberos.examples.getS4U2self import getS4U2self
	from minikerberos.examples.getTGT import getTGT
	try:
		await getTGT(KERBEROS_CONN_URL_MACHINE, ccachefile='test.ccache')
		with tempfile.NamedTemporaryFile() as d:
			with tempfile.NamedTemporaryFile() as f:
				await getS4U2self(KERBEROS_CONN_URL_MACHINE_CCACHE, KERBEROS_DELEGATION_SPN_SELF, KERBEROS_DELEGATION_USER_SELF,d.name, f.name)
				ccache = CCACHE.from_file(f.name)
				assert len(ccache.get_hashes()) > 0
			kirbi = Kirbi.from_file(d.name)
			assert kirbi.kirbiobj is not None
	finally:
		os.remove('test.ccache')

@pytest.mark.asyncio
async def test_getS4U2proxy():
	from minikerberos.examples.getS4U2proxy import getS4U2proxy
	with tempfile.NamedTemporaryFile() as d:
		with tempfile.NamedTemporaryFile() as f:
			await getS4U2proxy(KERBEROS_CONN_URL_MACHINE, KERBEROS_DELEGATION_SPN_PROXY, KERBEROS_DELEGATION_USER_PROXY, d.name, f.name)
			ccache = CCACHE.from_file(f.name)
			assert len(ccache.get_hashes()) > 0
		kirbi = Kirbi.from_file(d.name)
		assert kirbi.kirbiobj is not None

@pytest.mark.asyncio
async def test_getS4U2proxy_ccache():
	import os
	from minikerberos.examples.getS4U2proxy import getS4U2proxy
	from minikerberos.examples.getTGT import getTGT
	try:
		await getTGT(KERBEROS_CONN_URL_MACHINE, ccachefile='test.ccache')
		with tempfile.NamedTemporaryFile() as d:
			with tempfile.NamedTemporaryFile() as f:
				await getS4U2proxy(KERBEROS_CONN_URL_MACHINE_CCACHE, KERBEROS_DELEGATION_SPN_PROXY, KERBEROS_DELEGATION_USER_PROXY,d.name, f.name)
				ccache = CCACHE.from_file(f.name)
				assert len(ccache.get_hashes()) > 0
			kirbi = Kirbi.from_file(d.name)
			assert kirbi.kirbiobj is not None
	finally:
		os.remove('test.ccache')

@pytest.mark.asyncio
async def test_getnt():
	from minikerberos.examples.getNT import get_NT
	#moving certificate to current path...
	current_file_path = pathlib.Path(__file__).parent.absolute()
	cert_file_path = current_file_path.joinpath('testdata', 'test.pfx')
	try:
		shutil.copy(cert_file_path, 'a.pfx')
		results = await get_NT(KERBEROS_CONN_URL_PKINIT_PFX)
		assert len(results) > 0
	finally:
		os.remove('a.pfx')

@pytest.mark.asyncio
async def test_cve_2022_33679():
	from minikerberos.examples.CVE_2022_33679 import exploit
	with tempfile.NamedTemporaryFile() as f:
		await exploit(KERBEROS_CONN_URL_ASREP_MD4ARC4, f.name)
		kirbi = Kirbi.from_file(f.name)
		assert kirbi.kirbiobj is not None




if __name__ == '__main__':
	asyncio.run(test_spnroast_base())



# Kudos:
# Parts of this code was inspired by the following project by @rubin_mor
# https://github.com/morRubin/AzureADJoinedMachinePTC
# 


from ctypes import GetLastError
from minikerberos.common.windows.defines import *

CMSG_DATA = 1
CMSG_SIGNED =  2
CMSG_ENVELOPED =  3
CMSG_SIGNED_AND_ENVELOPED = 4
CMSG_HASHED = 5

CMSG_TYPE_PARAM = 1
CMSG_CONTENT_PARAM = 2
CMSG_BARE_CONTENT_PARAM = 3
CMSG_INNER_CONTENT_TYPE_PARAM = 4
CMSG_SIGNER_COUNT_PARAM = 5
CMSG_SIGNER_INFO_PARAM = 6
CMSG_SIGNER_CERT_INFO_PARAM = 7
CMSG_SIGNER_HASH_ALGORITHM_PARAM = 8
CMSG_SIGNER_AUTH_ATTR_PARAM = 9
CMSG_SIGNER_UNAUTH_ATTR_PARAM = 10
CMSG_CERT_COUNT_PARAM = 11
CMSG_CERT_PARAM = 12
CMSG_CRL_COUNT_PARAM = 13
CMSG_CRL_PARAM = 14
CMSG_ENVELOPE_ALGORITHM_PARAM = 15
CMSG_RECIPIENT_COUNT_PARAM = 17
CMSG_RECIPIENT_INDEX_PARAM = 18
CMSG_RECIPIENT_INFO_PARAM = 19
CMSG_HASH_ALGORITHM_PARAM = 20
CMSG_HASH_DATA_PARAM = 21
CMSG_COMPUTED_HASH_PARAM = 22
CMSG_ENCRYPT_PARAM = 26


X509_ASN_ENCODING = 0x00000001
X509_NDR_ENCODING = 0x00000002
PKCS_7_ASN_ENCODING = 0x00010000
PKCS_7_NDR_ENCODING = 0x00020000

CMSG_CMS_ENCAPSULATED_CONTENT_FLAG = 0x00000040

CRYPT_ACQUIRE_CACHE_FLAG = 0x00000001
CRYPT_ACQUIRE_USE_PROV_INFO_FLAG = 0x00000002
CRYPT_ACQUIRE_COMPARE_KEY_FLAG = 0x00000004
CRYPT_ACQUIRE_NO_HEALING = 0x00000008
CRYPT_ACQUIRE_SILENT_FLAG = 0x00000040
CRYPT_ACQUIRE_NCRYPT_KEY_FLAGS_MASK = 0x00070000
CRYPT_ACQUIRE_ALLOW_NCRYPT_KEY_FLAG = 0x00010000
CRYPT_ACQUIRE_PREFER_NCRYPT_KEY_FLAG = 0x00020000
CRYPT_ACQUIRE_ONLY_NCRYPT_KEY_FLAG = 0x00040000

HCRYPTMSG = HANDLE
HCERTSTORE = HANDLE
HCRYPTPROV_OR_NCRYPT_KEY_HANDLE = HANDLE
PHCRYPTPROV_OR_NCRYPT_KEY_HANDLE = POINTER(HCRYPTPROV_OR_NCRYPT_KEY_HANDLE)

class FILETIME(Structure):
	_fields_ = [
		('dwLowDateTime', DWORD),
		('dwHighDateTime', DWORD),
	]


class CRYPTOAPI_BLOB(Structure):
	_fields_ = [
		('cbData', DWORD),
		('pbData', LPBYTE),
	]

	@staticmethod
	def get_empty():
		t = CRYPTOAPI_BLOB()
		t.cbData = 0
		t.pbData = None
		return t

CRYPT_INTEGER_BLOB = CRYPTOAPI_BLOB
CERT_NAME_BLOB = CRYPTOAPI_BLOB
CRL_BLOB = CRYPTOAPI_BLOB
PCRL_BLOB = POINTER(CRL_BLOB)
CERT_BLOB = CRYPTOAPI_BLOB
PCERT_BLOB = POINTER(CERT_BLOB)
CRYPT_BIT_BLOB = CRYPTOAPI_BLOB
CRYPT_OBJID_BLOB = CRYPTOAPI_BLOB
CRYPT_HASH_BLOB = CRYPTOAPI_BLOB
PCRYPT_ATTR_BLOB = POINTER(CRYPTOAPI_BLOB)

class CRYPT_ALGORITHM_IDENTIFIER(Structure):
	_fields_ = [
		('pszObjId', LPSTR),
		('Parameters', CRYPT_OBJID_BLOB),
	]

class CRYPT_ATTRIBUTE(Structure):
	_fields_ = [
		("pszObjId", LPSTR),
		("cValue",   DWORD),
		("rgValue",  PCRYPT_ATTR_BLOB),
	]
PCRYPT_ATTRIBUTE = POINTER(CRYPT_ATTRIBUTE)

class CERT_PUBLIC_KEY_INFO(Structure):
	_fields_ = [
		('Algorithm', CRYPT_ALGORITHM_IDENTIFIER),
		('PublicKey', CRYPT_BIT_BLOB),
	]

class CERT_EXTENSION(Structure):
	_fields_ = [
		('pszObjId', LPSTR),
		('fCritical', BOOL),
		('Value', CRYPT_OBJID_BLOB),
	]

class CERT_ISSUER_SERIAL_NUMBER(Structure):
	_fields_ = [
		("Issuer", CERT_NAME_BLOB),
		("SerialNumber", CRYPT_INTEGER_BLOB),
	]
PCERT_ISSUER_SERIAL_NUMBER = POINTER(CERT_ISSUER_SERIAL_NUMBER)

class TMPUNION_CERT_ID(Union):
	_fields_ = [
		("IssuerSerialNumber", CERT_ISSUER_SERIAL_NUMBER),
		("KeyId", CRYPT_HASH_BLOB),
		("HashId", CRYPT_HASH_BLOB),
	]
PTMPUNION_CERT_ID = POINTER(TMPUNION_CERT_ID)

class CERT_ID(Structure):
	_fields_ = [
		("dwIdChoice",     DWORD),
		("DUMMYUNIONNAME", TMPUNION_CERT_ID),
	]
PCERT_ID = POINTER(CERT_ID)

PCERT_EXTENSION  = POINTER(CERT_EXTENSION)

class CERT_INFO(Structure):
	_fields_ = [
		('dwVersion',          DWORD),
		('SerialNumber',       CRYPT_INTEGER_BLOB),
		('SignatureAlgorithm', CRYPT_ALGORITHM_IDENTIFIER),
		('Issuer',             CERT_NAME_BLOB),
		('NotBefore',          FILETIME),
		('NotAfter',           FILETIME),
		('Subject',            CERT_NAME_BLOB),
		('SubjectPublicKeyInfo',  CERT_PUBLIC_KEY_INFO),
		('IssuerUniqueId',     CRYPT_BIT_BLOB),
		('SubjectUniqueId',    CRYPT_BIT_BLOB),
		('cExtension',         DWORD),
		('rgExtension',        PCERT_EXTENSION),
	]
PCERT_INFO  = POINTER(CERT_INFO)


HCRYPTPROV = HANDLE
NCRYPT_KEY_HANDLE = HANDLE

class CMSG_SIGNER_ENCODE_INFO_UNION(Union):
	_fields_ = [
		("hCryptProv", HCRYPTPROV),
		("hNCryptKey", NCRYPT_KEY_HANDLE)
]
PCMSG_SIGNER_ENCODE_INFO_UNION = POINTER(CMSG_SIGNER_ENCODE_INFO_UNION)

class DUMMYUNIONNAME(ctypes.Union):
	_fields_ = [
		("hCryptProv", HCRYPTPROV),
		("hNCryptKey", NCRYPT_KEY_HANDLE)]

class CMSG_SIGNER_ENCODE_INFO(Structure):
	#_anonymous_ = ("u",)
	_fields_ = [
		('cbSize',        DWORD),
		('pCertInfo',     PCERT_INFO),
		('DUMMYUNIONNAME', LPVOID), # CMSG_SIGNER_ENCODE_INFO_UNION
		('dwKeySpec',     DWORD),
		('HashAlgorithm', CRYPT_ALGORITHM_IDENTIFIER ),
		('pvHashAuxInfo', PVOID),
		('cAuthAttr',     DWORD),
		('rgAuthAttr',    PCRYPT_ATTRIBUTE),
		('cUnauthAttr',   DWORD),
		('rgUnauthAttr',  PCRYPT_ATTRIBUTE),
		('SignerId',      CERT_ID),
		#('HashEncryptionAlgorithm', CRYPT_ALGORITHM_IDENTIFIER),
		#('pvHashEncryptionAuxInfo', PVOID),
	]
PCMSG_SIGNER_ENCODE_INFO  = POINTER(CMSG_SIGNER_ENCODE_INFO)


class CMSG_SIGNED_ENCODE_INFO(Structure):
	_fields_ = [
		("cbSize", DWORD),
		("cSigners", DWORD),
		("rgSigners", PCMSG_SIGNER_ENCODE_INFO),
		("cCertEncoded", DWORD),
		("rgCertEncoded", PCERT_BLOB),
		("cCrlEncoded", DWORD),
		("rgCrlEncoded", PCRL_BLOB),
		("cAttrCertEncoded", DWORD),
		("rgAttrCertEncoded", PCERT_BLOB),
	
	]
PCMSG_SIGNED_ENCODE_INFO = POINTER(CMSG_SIGNED_ENCODE_INFO)
PPCMSG_SIGNED_ENCODE_INFO = POINTER(PCMSG_SIGNED_ENCODE_INFO)

class CERT_CONTEXT(Structure):
	_fields_ = [
		('dwCertEncodingType', DWORD),
		('pbCertEncoded',      LPBYTE),
		('cbCertEncoded',      DWORD),
		('pCertInfo',          PCERT_INFO),
		('hCertStore',         HCERTSTORE),
	]
PCERT_CONTEXT  = POINTER(CERT_CONTEXT)
PCCERT_CONTEXT = PCERT_CONTEXT


class CMSG_STREAM_INFO(Structure):
	_fields_ = [
		("cbContent",       DWORD),
		("pfnStreamOutput", PVOID),
		("pvArg",           PVOID),
	]
PCMSG_STREAM_INFO = POINTER(CMSG_STREAM_INFO)


# https://docs.microsoft.com/en-us/windows/win32/api/wincrypt/nf-wincrypt-certopensystemstorew
def CertOpenSystemStore(subSystemProtocol = 'MY'):
	_CertOpenSystemStore = windll.crypt32.CertOpenSystemStoreW
	_CertOpenSystemStore.argtypes = [PVOID, LPWSTR]
	_CertOpenSystemStore.restype = HCERTSTORE
	_CertOpenSystemStore.errcheck = RaiseIfZero
	
	handle = _CertOpenSystemStore(None, subSystemProtocol)

	return handle

# https://docs.microsoft.com/en-us/windows/win32/api/wincrypt/nf-wincrypt-certclosestore
def CertCloseStore(handle, dwFlags = 0):
	_CertCloseStore = windll.crypt32.CertCloseStore
	_CertCloseStore.argtypes = [HCERTSTORE, DWORD]
	_CertCloseStore.restype = BOOL
	_CertCloseStore.errcheck = RaiseIfZero
	
	handle = _CertCloseStore(handle, dwFlags)

	return handle

def CertFreeCertificateContext(handle):
	_CertFreeCertificateContext = windll.crypt32.CertFreeCertificateContext
	_CertFreeCertificateContext.argtypes = [PCCERT_CONTEXT]
	_CertFreeCertificateContext.restype = BOOL
	
	_CertFreeCertificateContext(handle)

def CertEnumCertificatesInStore(handle, prev_cert_handle = None):
	_CertEnumCertificatesInStore = windll.crypt32.CertEnumCertificatesInStore
	_CertEnumCertificatesInStore.argtypes = [HCERTSTORE, PCCERT_CONTEXT]
	_CertEnumCertificatesInStore.restype = PCCERT_CONTEXT
	#_CertEnumCertificatesInStore.errcheck = RaiseIfZero
	
	handle = _CertEnumCertificatesInStore(handle, prev_cert_handle)

	return handle

def CryptAcquireCertificatePrivateKey(handle, dwFlags = CRYPT_ACQUIRE_ALLOW_NCRYPT_KEY_FLAG):
	_CryptAcquireCertificatePrivateKey = windll.crypt32.CryptAcquireCertificatePrivateKey
	_CryptAcquireCertificatePrivateKey.argtypes = [PCCERT_CONTEXT, DWORD, PVOID, PHCRYPTPROV_OR_NCRYPT_KEY_HANDLE, LPDWORD, LPBOOL]
	_CryptAcquireCertificatePrivateKey.restype = BOOL
	_CryptAcquireCertificatePrivateKey.errcheck = RaiseIfZero

	dwKeySpec = DWORD(0)
	fCallerFreeProvOrNCryptKey = BOOL(0)
	hprov = HANDLE(0)

	_CryptAcquireCertificatePrivateKey(handle, dwFlags, None, byref(hprov), byref(dwKeySpec),  byref(fCallerFreeProvOrNCryptKey))

	return hprov, dwKeySpec, fCallerFreeProvOrNCryptKey


def CryptMsgOpenToEncode(MsgEncodeInfo, dwMsgEncodingType = X509_ASN_ENCODING | PKCS_7_ASN_ENCODING, dwFlags = CMSG_CMS_ENCAPSULATED_CONTENT_FLAG, dwMsgType = CMSG_SIGNED, InnerContentObjID = b"1.3.6.1.5.2.3.1", StreamInfo = None):
	_CryptMsgOpenToEncode = windll.crypt32.CryptMsgOpenToEncode
	_CryptMsgOpenToEncode.argtypes = [DWORD, DWORD, DWORD, PCMSG_SIGNED_ENCODE_INFO, LPSTR, PCMSG_STREAM_INFO]
	_CryptMsgOpenToEncode.restype = HCRYPTMSG
	_CryptMsgOpenToEncode.errcheck = RaiseIfZero

	if isinstance(InnerContentObjID, str):
		InnerContentObjID = ctypes.create_string_buffer(InnerContentObjID)
	res = _CryptMsgOpenToEncode(dwMsgEncodingType, dwFlags, dwMsgType, byref(MsgEncodeInfo), InnerContentObjID, StreamInfo)

	return res

def CryptMsgUpdate(hCryptMsg, data, fFinal = True):
	_CryptMsgUpdate = windll.crypt32.CryptMsgUpdate
	_CryptMsgUpdate.argtypes = [HCRYPTMSG, PVOID, DWORD, BOOL]
	_CryptMsgUpdate.restype = BOOL
	_CryptMsgUpdate.errcheck = RaiseIfZero

	dlen = len(data)
	data = ctypes.create_string_buffer(data, len(data))
	res = _CryptMsgUpdate(hCryptMsg, data, dlen, fFinal)

	return res

def CryptMsgGetParam(hCryptMsg, ptype, dwIndex = 0):
	_CryptMsgGetParam = windll.crypt32.CryptMsgGetParam
	_CryptMsgGetParam.argtypes = [HCRYPTMSG, DWORD, DWORD, PVOID, LPDWORD]
	_CryptMsgGetParam.restype = BOOL

	
	dlen = DWORD(0)
	res = _CryptMsgGetParam(hCryptMsg, ptype, dwIndex, None, byref(dlen))
	
	data = ctypes.create_string_buffer(dlen.value)
	res = _CryptMsgGetParam(hCryptMsg, ptype, dwIndex, byref(data), byref(dlen))
	if res != True:
		raise ctypes.WinError(GetLastError())

	return data.raw


def get_cert(pccert, native = False):
	from asn1crypto.x509 import Certificate
	cctx = pccert.contents
	cert_data = ctypes.string_at(cctx.pbCertEncoded, cctx.cbCertEncoded)
	if native is False:
		return Certificate.load(cert_data)
	return Certificate.load(cert_data).native

def list_certstore(certstore_name = 'MY'):
	chandle = CertOpenSystemStore(certstore_name)
	hcert = None
	while True:
		hcert = CertEnumCertificatesInStore(chandle, hcert)
		if bool(hcert) is False:
			#null ptr means no more certs
			break

		certificate = get_cert(hcert, True)
		if 'tbs_certificate' in certificate:
			if 'subject' in certificate['tbs_certificate']:
				subject = certificate['tbs_certificate']['subject']#['common_name']
				if isinstance(subject, str):
					subject = [subject]
				print(subject)

		else:
			input('!')
	if bool(hcert) is True:
		CertFreeCertificateContext(hcert)

def find_cert_by_cn(common_name, certstore_name = 'MY'):
	chandle = CertOpenSystemStore(certstore_name)
	hcert = None
	while True:
		hcert = CertEnumCertificatesInStore(chandle, hcert)
		if bool(hcert) is False:
			raise Exception('Couldnt find certificate for %s in certstore %s' % (common_name, certstore_name))
		certificate = get_cert(hcert)
		subject = certificate.subject.native['common_name']
		if isinstance(subject, list):
			for se in subject:
				if se == common_name:
					return certificate, hcert, chandle
		else:
			if subject == common_name:
				return certificate, hcert, chandle

	

def pkcs7_sign(hcert, data):
	hprov, keyspec, to_free = CryptAcquireCertificatePrivateKey(hcert)

	hashalgo = CRYPT_ALGORITHM_IDENTIFIER()
	hashalgo.pszObjId = b"1.3.14.3.2.26" #szOID_OIWSEC_sha1
	hashalgo.Parameters = CRYPTOAPI_BLOB.get_empty()

	Signers = CMSG_SIGNER_ENCODE_INFO()
	Signers.cbSize = ctypes.sizeof(CMSG_SIGNER_ENCODE_INFO)
	Signers.pCertInfo = hcert.contents.pCertInfo
	Signers.DUMMYUNIONNAME = hprov
	Signers.dwKeySpec = keyspec
	Signers.HashAlgorithm = hashalgo
	Signers.pvHashAuxInfo = None
	Signers.cAuthAttr = 0
	Signers.rgAuthAttr = None
	Signers.cUnauthAttr = 0
	Signers.rgUnauthAttr = None

	Certificate = CERT_BLOB()
	Certificate.cbData = hcert.contents.cbCertEncoded
	Certificate.pbData = hcert.contents.pbCertEncoded

	MsgEncodeInfo = CMSG_SIGNED_ENCODE_INFO()
	MsgEncodeInfo.cbSize = ctypes.sizeof(CMSG_SIGNED_ENCODE_INFO)
	MsgEncodeInfo.cSigners = 1
	MsgEncodeInfo.rgSigners = ctypes.pointer(Signers)
	MsgEncodeInfo.cCertEncoded = 1
	MsgEncodeInfo.rgCertEncoded = ctypes.pointer(Certificate)
	MsgEncodeInfo.cCrlEncoded = 0
	MsgEncodeInfo.rgCrlEncoded = None
	MsgEncodeInfo.cAttrCertEncoded  = 0
	MsgEncodeInfo.rgAttrCertEncoded = None

	hmsg = CryptMsgOpenToEncode(MsgEncodeInfo)

	CryptMsgUpdate(hmsg, data, True)
	res = CryptMsgGetParam(hmsg, CMSG_CONTENT_PARAM)
	
	return res
